from comfit.models.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint

class PhaseFieldCrystal3DBodyCenteredCubic(PhaseFieldCrystal):
    def __init__(self, nx, ny, nz, **kwargs):
        """Initializes a phase field crystal system in 3D with a body centered cubic crystal structure.

        Args:
            nx: The number of unit cells in the x direction.
            ny: The number of unit cells in the y direction.
            nz: The number of unit cells in the z direction.

        Returns:
            The system object representing the PhaseFieldCrystal3DBodyCenteredCubic simulation.
        """

        # Type of the system
        self.type = 'PhaseFieldCrystal3DBodyCenteredCubic'
        self.dim = 3
        self.nx = nx
        self.ny = ny
        self.nz = nz

        # Default simulation parameters
        self.micro_resolution = [7, 7, 7]
        self.psi0 = -0.325
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]
        self.yRes = ny * self.micro_resolution[1]
        self.zRes = nz * self.micro_resolution[2]

        a0 = 2 * np.pi * np.sqrt(2)
        self.a = a0 / 2 * np.array([[-1, 1, 1],
                                    [1, -1, 1],
                                    [1, 1, -1],
                                    [1, 1, 1]])

        self.q = np.array([[0, 1, 1],
                           [1, 0, 1],
                           [1, 1, 0],
                           [0, -1, 1],
                           [-1, 0, 1],
                           [-1, 1, 0]]) / np.sqrt(2)

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 6
        self.number_of_primary_reciprocal_lattice_modes = 6

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]
        self.dy = a0 / self.micro_resolution[1]
        self.dz = a0 / self.micro_resolution[2]

        self.A = self.calc_proto_amplitudes_conserved()
        self.eta0 = np.array([self.A, self.A, self.A, self.A, self.A, self.A])

        # Set the elastic constants
        self.el_lambda = 4 * self.A ** 2
        self.el_mu = 4 * self.A ** 2
        self.el_gamma = - 4*self.A**2

        # Initialize the BaseSystem
        super().__init__(self.dim, xRes=self.xRes, yRes=self.yRes, zRes=self.zRes,
                         dx=self.dx, dy=self.dy, dz=self.dz, dt=self.dt)

        # Set the a0
        self.a0 = a0
        self.defined_length_scale = True


    def calc_proto_amplitudes_conserved(self):
        """Calculates the proto-amplitudes for the system.

        Args:
            None

        Returns:
            The proto-amplitudes for the system.
        """

        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v

        # Testing which of the three amplitudes give the lowest free energy
        A=0
        A_tmp = 1/(45*v)*( -2*t - 6*v*psi0 + np.sqrt(4*t**2 - 45*r*v - 66*t*v*psi0 - 99*v**2*psi0**2))
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        A_tmp = 1/(45*v)*( -2*t - 6*v*psi0 - np.sqrt(4*t**2 - 45*r*v - 66*t*v*psi0 - 99*v**2*psi0**2))
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        return A

    def calc_free_energy_from_proto_amplitudes(self, psi0, A):
        """Calculates the free energy of the phase-field crystal from the proto amplitudes.

        Args:
            psi0: The average value of psi.
            A: The proto-amplitude.

        Returns:
            The free energy of the phase-field crystal.
        """
        r = self.r
        t = self.t
        v = self.v

        return 4*np.sqrt(2)*np.pi**3/3*(1620*A**4*v + 192*A**3*(t + 3*v*psi0) + psi0**2*(6 + 6*r + 4*t*psi0 + 3*v*psi0**2) + 72*A**2*(r + psi0*(2*t + 3*v*psi0)))

    def calc_omega_f(self):
        """Calculates the linear part of the evolution equation in Fourier space.

        Args:
            None

        Returns:
            The linear part of the evolution equation in Fourier space.
        """

        k2 = self.calc_k2()
        return - k2 * (self.r + (1 - k2)**2)

    def calc_stress_tensor_microscopic(self):
        """Calculates the microscopic stress of the phase-field crystal.

        Args:
            None
        
        Returns:
            The microscopic stress of the phase-field crystal.
        """
        stress = np.zeros((6,self.xRes,self.yRes,self.zRes))
        
        L1psi = np.real(sp.fft.ifftn((1-self.calc_k2())*self.psi_f))
        stress[0] = -2*L1psi*np.real(sp.fft.ifftn(self.dif[0]*self.dif[0]*self.psi_f))
        stress[1] = -2*L1psi*np.real(sp.fft.ifftn(self.dif[0]*self.dif[1]*self.psi_f))
        stress[2] = -2*L1psi*np.real(sp.fft.ifftn(self.dif[0]*self.dif[2]*self.psi_f))
        stress[3] = -2*L1psi*np.real(sp.fft.ifftn(self.dif[1]*self.dif[1]*self.psi_f))
        stress[4] = -2*L1psi*np.real(sp.fft.ifftn(self.dif[1]*self.dif[2]*self.psi_f))
        stress[5] = -2*L1psi*np.real(sp.fft.ifftn(self.dif[2]*self.dif[2]*self.psi_f))

        return stress
    
    def calc_stress_divergence_f(self, field_f = None):
        """Calculates the divergence of the stress tensor in Fourier space.

        Args:
            field_f: The field in Fourier space.

        Returns:
            The divergence of the stress tensor in Fourier space.
        """
        
        if field_f is None:
            field_f = self.psi_f
        
        k2 = self.calc_k2()

        return np.array([
            -2*self.calc_Gaussian_filter_f()*sp.fft.fftn(sum([
                sp.fft.ifftn((1-k2)*self.dif[i]*field_f)*sp.fft.ifftn(self.dif[i]*self.dif[j]*field_f) for i in range(self.dim)
                ]) + sp.fft.ifftn((1-k2)*field_f)*sp.fft.ifftn(self.dif[j]*(-k2)*field_f)) for j in range(self.dim)]
                )


import numpy as np
import scipy as sp

class BaseSystemInit:
    """ Initialization methods for the base system class"""
    def __init__(self, dim, **kwargs):
        """
        Initialize the class with the given parameters.

        Args:
            dimension (int): The dimension of the system. Must be 1, 2, or 3.
            **kwargs: Additional keyword arguments, see 
            https://comfitlib.com/ClassBaseSystem/

        Returns:
            None
        """
        # Set default plotting library
        self.plot_lib = 'matplotlib'

        if dim not in [1, 2, 3]:
            raise ValueError('Dimension must be 1, 2, or 3.')

        self.dim = dim

        if 'xlim' in kwargs:
            # Providing xlim trumps providing xmin and xmax
            self.xmin = kwargs['xlim'][0]
            self.xmax = kwargs['xlim'][1]
            
            self.xRes = kwargs.get('xRes', 101)
            self.dx = (self.xmax - self.xmin) / self.xRes
        
        else:
            self.xmin = kwargs.get('xmin', 0)
            self.xRes = kwargs.get('xRes', 101)

            self.dx = kwargs.get('dx', 1.0)

            if 'xmax' in kwargs:
                self.xmax = kwargs['xmax']
            else:
                self.xmax = self.xmin + self.xRes * self.dx

        self.xlim = [self.xmin, self.xmax]

        # Providing dx trumps providing xRes
        if 'dx' in kwargs:
            self.dx = kwargs['dx']
            self.xRes = round((self.xmax - self.xmin) / self.dx)

        # Setting default values for y and z for 1 dimensional systems
        self.ymin = kwargs.get('ymin', 0)
        self.zmin = kwargs.get('zmin', 0)
        
        self.dy = kwargs.get('dy', 1.0)
        self.dz = kwargs.get('dz', 1.0)

        self.ymax = kwargs.get('ymax',1)
        self.zmax = kwargs.get('zmax',1)

        self.yRes = kwargs.get('yRes', 1)
        self.zRes = kwargs.get('zRes', 1)

        if self.dim > 1:
            if 'ylim' in kwargs:
                self.ymin = kwargs['ylim'][0]
                self.ymax = kwargs['ylim'][1]
                
                self.yRes = kwargs.get('yRes', 101)
                self.dy = (self.ymax - self.ymin) / self.yRes
            
            else:
                self.ymin = kwargs.get('ymin', 0)
                self.yRes = kwargs.get('yRes', 101)
                self.dy = kwargs.get('dy', 1.0)

                if 'ymax' in kwargs:
                    self.ymax = kwargs['ymax']
                else:
                    self.ymax = self.ymin + self.yRes * self.dy

            # Providing dy trumps providing yRes
            if 'dy' in kwargs:
                self.dy = kwargs['dy']
                self.yRes = round((self.ymax - self.ymin) / self.dy)
        
        if self.dim > 2:
            if 'zlim' in kwargs:
                self.zmin = kwargs['zlim'][0]
                self.zmax = kwargs['zlim'][1]
                
                self.zRes = kwargs.get('zRes', 101)
                self.dz = (self.zmax - self.zmin) / self.zRes

            else:
                self.zmin = kwargs.get('zmin', 0)
                self.zRes = kwargs.get('zRes', 101)
                self.dz = kwargs.get('dz', 1.0)

                if 'zmax' in kwargs:
                    self.zmax = kwargs['zmax']
                else:
                    self.zmax = self.zmin + self.zRes * self.dz

            # Providing dz trumps providing zRes
            if 'dz' in kwargs:
                self.dz = kwargs['dz']
                self.zRes = round((self.zmax - self.zmin) / self.dz)

        # Minimum and maximum sizes of the domain
        if self.dim == 1:
            self.Lmin = self.xmax - self.xmin
            self.Lmax = self.xmax - self.xmin
        elif self.dim == 2:
            self.Lmin = np.min([self.xmax - self.xmin, self.ymax - self.ymin])
            self.Lmax = np.max([self.xmax - self.xmin, self.ymax - self.ymin])
        elif self.dim == 3:
            self.Lmin = np.min([self.xmax - self.xmin, self.ymax - self.ymin, self.zmax - self.zmin])
            self.Lmax = np.max([self.xmax - self.xmin, self.ymax - self.ymin, self.zmax - self.zmin])

        #  Setting default value of time
        self.time = kwargs.get('time', 0)
        self.dt = kwargs.get('dt', 0.1)

        # Construct parameters
        self.x = np.linspace(self.xmin, self.xmax-self.dx, self.xRes)
        self.y = np.linspace(self.ymin, self.ymax-self.dy, self.yRes)
        self.z = np.linspace(self.zmin, self.zmax-self.dz, self.zRes)

        self.Res = self.xRes * self.yRes * self.zRes
        if self.dim == 1:
            self.dims = self.xRes
        elif self.dim == 2:
            self.dims = [self.xRes, self.yRes]
        elif self.dim == 3:
            self.dims = [self.xRes, self.yRes, self.zRes]

        self.a0 = 1  # System length scale

        # Helpful midpoints and their indices
        self.xmidi = (self.xRes) // 2 
        self.xmid = self.x[self.xmidi]

        self.ymidi = (self.yRes) // 2 
        self.ymid = self.y[self.ymidi]

        self.zmidi = (self.zRes) // 2 
        self.zmid = self.z[self.zmidi]

        self.midi = self.xRes * self.yRes * (self.zmidi - 1) + self.yRes * (self.xmidi - 1) + self.ymidi
        if self.dim == 1:
            self.rmid = self.xmid
            self.zero_index = 0 #Nothing comment
        elif self.dim == 2:
            self.rmid = [self.xmid, self.ymid]
            self.zero_index = (0,0)
        elif self.dim == 3:
            self.rmid = [self.xmid, self.ymid, self.zmid]
            self.zero_index = (0,0,0)

        # Fourier modes
        self.k = [self.calc_wavenums(self.x)]
        if self.dim == 2:
            self.k[0] = self.k[0].reshape(self.xRes, 1)
            self.k.append(self.calc_wavenums(self.y).reshape(1, self.yRes))
        elif self.dim == 3:
            self.k[0] = self.k[0].reshape(self.xRes, 1, 1)
            self.k.append(self.calc_wavenums(self.y).reshape(1, self.yRes, 1))
            self.k.append(self.calc_wavenums(self.z).reshape(1, 1, self.zRes))

        # Reshape the position vectors
        if self.dim == 2:
            self.x = self.x.reshape((self.xRes, 1))
            self.y = self.y.reshape((1, self.yRes))
        elif self.dim == 3:
            self.x = self.x.reshape((self.xRes, 1, 1))
            self.y = self.y.reshape((1, self.yRes, 1))
            self.z = self.z.reshape((1, 1, self.zRes))


        # Derivatives
        self.dif = [1j * ki for ki in self.k]

        self.dV = self.dx
        if self.dim > 1:
            self.dV *= self.dy
        if self.dim > 2:
            self.dV *= self.dz

        self.rmin = [self.xmin, self.ymin, self.zmin]
        self.rmax = [self.xmax, self.ymax, self.zmax]

    def __str__(self):
        """Return a string representation of the class.
        
        Input:
            None

        Returns:
            A string representation of the class.
        """
        description = f"BaseSystem instance\n"
        # Start with the dimension of the system
        description = f"System Dimension: {self.dim}\n"

        # Add x-axis properties
        description += f"X-Axis Limits: [{self.xmin}, {self.xmax}], Resolution: {self.xRes}, Delta: {self.dx}\n"

        # Add y-axis properties if dim > 1
        if self.dim > 1:
            description += f"Y-Axis Limits: [{self.ymin}, {self.ymax}], Resolution: {self.yRes}, Delta: {self.dy}\n"

        # Add z-axis properties if dim > 2
        if self.dim > 2:
            description += f"Z-Axis Limits: [{self.zmin}, {self.zmax}], Resolution: {self.zRes}, Delta: {self.dz}\n"

        # Add time properties
        description += f"Current Time: {self.time}, Time Step: {self.dt}"

        return description




from comfit.models.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint

class PhaseFieldCrystal2DTriangular(PhaseFieldCrystal):
    def __init__(self, nx, ny, **kwargs):
        """Initializes a phase field crystal system in 2D with a triangular crystal structure.

        Args:
            nx: The number of unit cells in the x direction.
            ny: The number of unit cells in the y direction.

        Returns:
            The system object representing the PhaseFieldCrystal2DTriangular simulation.
        """

        # Type of the system
        self.type = 'PhaseFieldCrystal2DTriangular'
        self.dim = 2
        self.nx = nx
        self.ny = ny

        # Default simulation parameters
        self.micro_resolution = [7, 12]
        self.psi0 = -0.3
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]
        self.yRes = ny * self.micro_resolution[1]

        a0 = 2 * np.pi * 2 / np.sqrt(3)
        self.a = a0 * np.array([[1, 0], [1 / 2, np.sqrt(3) / 2], [1 / 2, -np.sqrt(3) / 2]])
        self.q = np.array([[np.sqrt(3) / 2, -1 / 2], [0, 1], [-np.sqrt(3) / 2, -1 / 2]])

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 3
        self.number_of_primary_reciprocal_lattice_modes = 3

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]
        self.dy = np.sqrt(3) * a0 / self.micro_resolution[1]

        self.A = self.calc_proto_amplitudes_conserved()
        self.eta0 = np.array([self.A, self.A, self.A])

        # Set the elastic constants
        self.el_lambda = 3 * self.A ** 2
        self.el_mu = 3 * self.A ** 2
        self.el_gamma = 0


        # Initialize the BaseSystem
        super().__init__(self.dim, xRes=self.xRes, yRes=self.yRes,
                         dx=self.dx, dy=self.dy, dt=self.dt)

        # Set the a0
        self.a0 = a0
        self.defined_length_scale = True

        

    def calc_proto_amplitudes_conserved(self):
        """Calculates the proto-amplitudes for the system.

        Args:
            None
        
        Returns:
            The proto-amplitudes for the system.
        """

        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v
        
        #Testing which of the three amplitudes give the lowest free energy
        A = 0
        A_tmp = 1 / (15 * v)*(-t -3 * v * psi0 - np.sqrt(t**2 - 15 * v * r - 24 * t * v * psi0 - 36 * v ** 2 * psi0 ** 2)) 
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        A_tmp = 1 / (15 * v)*(-t -3 * v * psi0 + np.sqrt(t**2 - 15 * v * r - 24 * t * v * psi0 - 36 * v ** 2 * psi0 ** 2)) 
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        return A

    def calc_free_energy_from_proto_amplitudes(self, psi0, A):
        """Calculates the free energy of the system from the proto-amplitudes.

        Args:
            psi0: The average value of psi.
            A: The proto-amplitude.

        Returns:
            The free energy of the system.
        """
        r = self.r
        t = self.t
        v = self.v

        return np.pi**2/(3*np.sqrt(3))*(270*A**4*v + 48*A**3*( t + 3*v*psi0) + psi0**2*(6 + 6*r + 4*t*psi0 + 3*v*psi0**2) + 36*A**2*( r + psi0*(2*t + 3*v*psi0)))

    def calc_omega_f(self):
        """Calculates the linear evolution operator of the system.

        Args:
            None
        
        Returns:
            The linear evolution operator of the system.
        """

        k2 = self.calc_k2()
        return - k2 * (self.r + (1 - k2)**2)
    
    def calc_stress_tensor_microscopic(self):
        """Calculates the microscopic stress of the phase-field crystal.

        Args:
            None

        Returns:
            The microscopic stress of the phase-field crystal.
        """
        stress = np.zeros((3,self.xRes,self.yRes))
        
        Lpsi = np.real(sp.fft.ifftn((1-self.calc_k2())*self.psi_f))
        stress[0] = -2*Lpsi*np.real(sp.fft.ifftn(self.dif[0]*self.dif[0]*self.psi_f))
        stress[1] = -2*Lpsi*np.real(sp.fft.ifftn(self.dif[0]*self.dif[1]*self.psi_f))
        stress[2] = -2*Lpsi*np.real(sp.fft.ifftn(self.dif[1]*self.dif[1]*self.psi_f))

        return stress

    
    def calc_stress_divergence_f(self, field_f = None):
        """Calculates the divergence of the stress tensor in Fourier space.

        Args:
            field_f: The field in Fourier space.

        Returns:
            The divergence of the stress tensor in Fourier space.
        """

        if field_f is None:
            field_f = self.psi_f
        
        k2 = self.calc_k2()

        return np.array([
            -2*self.calc_Gaussian_filter_f()*sp.fft.fftn(sum([
                sp.fft.ifftn((1-k2)*self.dif[i]*field_f)*sp.fft.ifftn(self.dif[i]*self.dif[j]*field_f) for i in range(self.dim)
                ]) + sp.fft.ifftn((1-k2)*field_f)*sp.fft.ifftn(self.dif[j]*(-k2)*field_f)) for j in range(self.dim)]
                )


from .tool_colormaps import tool_colormap_bluewhitered, tool_colormap_angle, tool_colormap_sunburst
from .tool_animation import tool_save_plot, tool_make_animation_movie, tool_make_animation_gif
from .tool_export_plot_functions import tool_export_rotating_plot
from .tool_plot_manipulation_functions import tool_zoom_plot
from .tool_math_functions import tool_multinom


from comfit.core.base_system_init import BaseSystemInit
from comfit.core.base_system_conf import BaseSystemConf
from comfit.core.base_system_evolve import BaseSystemEvolve
from comfit.core.base_system_calc import BaseSystemCalc
from comfit.core.base_system_plot import BaseSystemPlot
from comfit.core.base_system_get import BaseSystemGet


class BaseSystem(BaseSystemInit, BaseSystemConf, BaseSystemEvolve, BaseSystemCalc, BaseSystemPlot, BaseSystemGet):
    """
    The BaseSystem class is the base class for all systems in ComFiT.
    """
    pass


import numpy as np
import scipy as sp

from comfit.tools.tool_colormaps import tool_colormap_angle, tool_colormap_bluewhitered, tool_colormap_sunburst
from comfit.tools.tool_create_orthonormal_triad import tool_create_orthonormal_triad

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.tri as mtri
import matplotlib.cm as cm

import plotly.graph_objects as go
import plotly.figure_factory as ff
import plotly.colors as pc

from skimage.measure import marching_cubes

import time

class BaseSystemPlot:
    """ Plotting methods for the base system class"""

    def show(self):
        """Show the plot."""
        if self.plot_lib == 'matplotlib':
            plt.show()
        elif self.plot_lib == 'plotly':
            go.show() #or fig.show?

    def plot_tool_set_axis_properties(self, **kwargs):
        """Sets the properties of the axis for a plot.
        
        Args:
            kwargs: keyword arguments for the axis properties

        Returns:
            The axis object with the properties set.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)
        
        if plot_lib == 'matplotlib':
            ax = kwargs.get('ax', plt.gca())
        elif plot_lib == 'plotly':
            fig = kwargs.get('fig', go.Figure())

        xlabel = kwargs.get('xlabel', 'x/a₀')

        if 'title' in kwargs:
                ax.set_title(kwargs['title'])

        #TODO: Check if I should instead include a fig object somewhere (Vidar 05.02.24)
        suptitle = kwargs.get('suptitle', None)
        if suptitle is not None:
            if plot_lib == 'matplotlib':
                ax.get_figure().suptitle(suptitle)
            elif plot_lib == 'plotly':
                fig.update_layout(title_text=suptitle)
        
        grid = kwargs.get('grid', True)
        if plot_lib == 'matplotlib':
            ax.grid(grid)

        xlim = [self.xmin/self.a0, (self.xmax-self.dx)/self.a0]

        if 'xmin' in kwargs:
            xlim[0] = kwargs['xmin'] / self.a0

        if 'xmax' in kwargs:
            xlim[1] = kwargs['xmax'] / self.a0

        if 'xlim' in kwargs:
            xlim = np.array(kwargs['xlim']) / self.a0

        ylim = [self.ymin/self.a0, (self.ymax-self.dy)/self.a0] if self.dim > 1 else None

        if 'ymin' in kwargs:
            ylim[0] = kwargs['ymin'] / self.a0 if self.dim > 1 else kwargs['ymin']
        
        if 'ymax' in kwargs:
            ylim[1] = kwargs['ymax'] / self.a0 if self.dim > 1 else kwargs['ymax']
        
        if 'ylim' in kwargs:
            ylim = np.array(kwargs['ylim'])/self.a0 if self.dim > 1 else kwargs['ylim']

        zlim = [self.zmin/self.a0, (self.zmax-self.dz)/self.a0] if self.dim > 2 else None

        if 'zmin' in kwargs:
             zlim[0] = kwargs['zmin'] / self.a0 if self.dim > 2 else kwargs['zmin']

        if 'zmax' in kwargs:
            zlim[1] = kwargs['zmax'] / self.a0 if self.dim > 2 else kwargs['zmax']

        if 'zlim' in kwargs:
            zlim = np.array(kwargs['zlim'])/self.a0 if self.dim > 2 else kwargs['zlim']

        # Custom x-ticks
        xticks = kwargs.get('xticks', None)
        if xticks is not None:
            if plot_lib == 'matplotlib':
                ax.set_xticks(xticks)
        
        xticklabels = kwargs.get('xticklabels', None)
        if xticklabels is not None:
            if plot_lib == 'matplotlib':
                ax.set_xticklabels(xticklabels)
            elif plot_lib == 'plotly':
                fig.update_layout(xaxis=dict(tickvals=xticks, ticktext=xticklabels))

        yticks = kwargs.get('yticks', None)
        if yticks is not None:
            if plot_lib == 'matplotlib':
                ax.set_yticks(yticks)
            elif plot_lib == 'plotly':
                fig.update_layout(yaxis=dict(tickvals=yticks))
        
        yticklabels = kwargs.get('yticklabels', None)
        if yticklabels is not None:
            if plot_lib == 'matplotlib':
                ax.set_yticklabels(yticklabels)
            elif plot_lib == 'plotly':
                fig.update_layout(yaxis=dict(ticktext=yticklabels))

        zticks = kwargs.get('zticks', None)
        if zticks is not None:
            if plot_lib == 'matplotlib':
                ax.set_zticks(zticks)
            elif plot_lib == 'plotly':
                fig.update_layout(zaxis=dict(tickvals=zticks))
        
        zticklabels = kwargs.get('zticklabels', None)
        if zticklabels is not None:
            if plot_lib == 'matplotlib':
                ax.set_zticklabels(zticklabels)
            elif plot_lib == 'plotly':
                fig.update_layout(zaxis=dict(ticktext=zticklabels))
        
        if self.dim == 1:
            # Set the xlabel
            if plot_lib == 'matplotlib':
                ax.set_xlabel(xlabel)
            elif plot_lib == 'plotly':
                fig.update_layout(xaxis_title=xlabel)

            # Set the ylabel
            if 'ylabel' in kwargs:
                if plot_lib == 'matplotlib':
                    ax.set_ylabel(kwargs['ylabel'])
                elif plot_lib == 'plotly':
                    fig.update_layout(yaxis_title=kwargs['ylabel'])

            # Set the title
            if 'title' in kwargs:
                if plot_lib == 'matplotlib':
                    ax.set_title(kwargs['title'])
                elif plot_lib == 'plotly':
                    fig.update_layout(title_text=kwargs['title'])
            
            if plot_lib == 'matplotlib':
                ax.set_xlim(xlim[0], xlim[1])
            elif plot_lib == 'plotly':
                fig.update_layout(xaxis_range=[xlim[0], xlim[1]])

            if ylim is not None:
                if plot_lib == 'matplotlib':
                    ax.set_ylim(ylim[0], ylim[1])
                elif plot_lib == 'plotly':
                    fig.update_layout(yaxis_range=[ylim[0], ylim[1]])

            if zlim is not None:
                if plot_lib == 'matplotlib':
                    ax.set_zlim(zlim[0], zlim[1])
                elif plot_lib == 'plotly':
                    fig.update_layout(zaxis_range=[zlim[0], zlim[1]])
        
        elif self.dim == 2:
            
            ylabel = kwargs.get('ylabel', 'y/a₀')

            if plot_lib == 'matplotlib':
                ax.set_xlabel(xlabel)
                ax.set_ylabel(ylabel)
            elif plot_lib == 'plotly':
                fig.update_layout(xaxis_title=xlabel)
                fig.update_layout(yaxis_title=ylabel)
            
            if plot_lib == 'matplotlib':
                ax.set_xlim(xlim[0], xlim[1])
                ax.set_ylim(ylim[0], ylim[1])
            elif plot_lib == 'plotly':
                fig.update_layout(xaxis_range=[xlim[0], xlim[1]])
                fig.update_layout(yaxis_range=[ylim[0], ylim[1]])

            if zlim is not None:
                if plot_lib == 'matplotlib':
                    ax.set_zlim(zlim[0], zlim[1])
                elif plot_lib == 'plotly':
                    fig.update_layout(zaxis_range=[zlim[0], zlim[1]])

            # Set the aspect ratio
            axis_equal = kwargs.get('axis_equal', True)
            if axis_equal:
                if plot_lib == 'matplotlib':
                    ax.set_aspect('equal')
                elif plot_lib == 'plotly':
                    fig.update_yaxes(
                        scaleanchor="x",
                        scaleratio=1,
                    )
                    pass

        elif self.dim == 3:

            ylabel = kwargs.get('ylabel', 'y/a₀')
            zlabel = kwargs.get('zlabel', 'z/a₀')

            if plot_lib == 'matplotlib':
                ax.set_xlabel(xlabel)
                ax.set_ylabel(ylabel)
                ax.set_zlabel(zlabel)

                ax.set_xlim3d(xlim[0], xlim[1])
                ax.set_ylim3d(ylim[0], ylim[1])
                ax.set_zlim3d(zlim[0], zlim[1])
            elif plot_lib == 'plotly':
                fig.update_layout(
                scene = dict(
                    xaxis=dict(title=xlabel, range=[xlim[0], xlim[1]]),
                    yaxis=dict(title=ylabel, range=[ylim[0], ylim[1]]),
                    zaxis=dict(title=zlabel, range=[zlim[0], zlim[1]])
                    )       
                    )

            elif plot_lib == 'plotly':
                fig.update_layout(xaxis_range=[xlim[0], xlim[1]])
                fig.update_layout(yaxis_range=[ylim[0], ylim[1]])
                fig.update_layout(zaxis_range=[zlim[0], zlim[1]])

            # Set the aspect ratio
            axis_equal = kwargs.get('axis_equal', True)
            if axis_equal:
                if plot_lib == 'matplotlib':
                    ax.set_aspect('equal')  
        if plot_lib == 'matplotlib':
            return ax
        elif plot_lib == 'plotly':
            return fig

    def plot_tool_surface_matplotlib(self, **kwargs):
        """Plots the surface of the given field.

        Args:
            **kwargs: Keyword arguments for the plot.
        
        Returns:
            The axes containing the plot.
        """
        field = kwargs['field']
        value = kwargs['value']
        ax = kwargs.get('ax', plt.gca())
        alpha = kwargs.get('alpha', 0.5)
        color = kwargs.get('color', 'b')

        verts, faces, _, _ = marching_cubes(field, value)
        x = (self.xmin+verts[:, 0]*self.dx)/self.a0
        y = (self.ymin+verts[:, 1]*self.dy)/self.a0
        z = (self.zmin+verts[:, 2]*self.dz)/self.a0
        ax.plot_trisurf(x, y, faces, z, alpha=alpha, color=color)

        return ax

    def plot_tool_extend_field(self,field):
        """Extends the field in case not a complete array is given. 

        For instance, if a field in 3 dimensions is calculated using only x, then the field is extended to the full 3D array.

        Args:
            field: The field to be extended

        Returns:
            The extended field (np.ndarray)
        """    
        print(self.yRes)

        # 2 dimensional fields
        if field.shape == (self.xRes,1):
            field = np.tile(field,(1,self.yRes))

        elif field.shape == (1,self.yRes):
            field = np.tile(field,(self.xRes,1))

        # 3 dimensional fields
        elif field.shape == (self.xRes,1,1):
            field = np.tile(field,(1,self.yRes,1))
            field = np.tile(field,(1,1,self.zRes))

        elif field.shape == (1,self.yRes,1):
            field = np.tile(field,(self.xRes,1,1))
            field = np.tile(field,(1,1,self.zRes))

        elif field.shape == (1,1,self.zRes):
            field = np.tile(field,(self.xRes,1,1))
            field = np.tile(field,(1,self.yRes,1))

        elif field.shape == (self.xRes,self.yRes,1):
            field = np.tile(field,(1,1,self.zRes))

        elif field.shape == (self.xRes,1,self.zRes):
            field = np.tile(field,(1,self.yRes,1))

        elif field.shape == (1,self.yRes,self.zRes):
            field = np.tile(field,(self.xRes,1,1))

        return field

    def plot_tool_plotly_add_angle_colorbar3D(self,fig):
        """Adds a colorbar to a 3D plot with the angle colormap.

        Args:
            fig: The figure to which the colorbar is added.
        
        Returns:
            The figure with the colorbar added.
        """
        # Generate the custom colormap
        custom_colormap = tool_colormap_angle(pyplot=True)

        fig.add_trace(go.Isosurface(
            x=[self.xmin/self.a0,self.xmax/self.a0], 
            y=[self.ymin/self.a0,self.ymax/self.a0], 
            z=[self.zmin/self.a0, self.zmax/self.a0],
            value=[0,0],
            cmin = -np.pi,
            cmax = np.pi,
            opacity=0,
            colorscale=custom_colormap,
            colorbar=dict(
                tickvals=[-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi],
                ticktext=['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
            ),
            showscale=True
        ))

        return fig


    def plot_field(self, field, **kwargs):
        """Plots the given (real) field.
        
        Args:
            field (array-like): The field to be plotted.
            **kwargs: Keyword arguments for the plot.
                See https://comfitlib.com/ClassBaseSystem/ 
                for a full list of keyword arguments.
        
        Returns:
            If self.plot_lib == 'matplotlib':
                The axes containing the plot (matplotlib.axes.Axes).
            If self.plot_lib == 'plotly':
                The figure containing the plot.
        """

        if field.dtype == bool:
            field = field.astype(float)

        # Check if the vector field is complex
        if np.iscomplexobj(field):
            print("\033[91mWarning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
            print('Max imaginary part: ', np.max(np.imag(field)))
            field = np.real(field)

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib=='matplotlib':
            # Check if an axis object is provided
            fig = kwargs.get('fig', plt.gcf())
            ax = kwargs.get('ax', None)

        elif plot_lib=='plotly':
            fig = kwargs.get('fig', go.Figure())

        # Kewyord arguments
        colorbar = kwargs.get('colorbar', True)
        axis_equal = kwargs.get('axis_equal',True)

        # Extend the field if not a complete array is given
        field = self.plot_tool_extend_field(field)
        
        if self.dim == 1:

            # Keyword arguments particular to the 1D case
            kwargs['grid'] = kwargs.get('grid', True)

            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111)

                ax.plot(self.x/self.a0, field)
            elif self.plot_lib == 'plotly':
                fig.add_trace(go.Scatter(
                    x=self.x/self.a0,
                    y=field,
                    mode='lines',
                    name='',
                    hovertemplate='x: %{x:.2f} a₀<br>field: %{y:.2f}'
                ))


        if self.dim == 2:
            
            # Keyword arguments particular to the 2D case
            kwargs['grid'] = kwargs.get('grid', False)

            if self.plot_lib == 'matplotlib':

                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111)
                
                # Set the colormap
                colormap = kwargs.get('colormap', 'viridis')

                if colormap == 'bluewhitered':
                    colormap = tool_colormap_bluewhitered()

                elif colormap == 'sunburst':
                    colormap = tool_colormap_sunburst()
                else:
                    colormap = plt.get_cmap(colormap)

                # Value limits symmetric
                vlim_symmetric = kwargs.get('vlim_symmetric', False)

                X, Y = np.meshgrid(self.x, self.y, indexing='ij')

                pcm = ax.pcolormesh(X / self.a0, Y / self.a0, field, shading='gouraud', cmap=colormap)

                xlim = [self.xmin, self.xmax-self.dx]
                ylim = [self.ymin, self.ymax-self.dy]

                limits_provided = False
                if 'xlim' in kwargs:
                    xlim = kwargs['xlim']
                    limits_provided = True
                else:
                    if 'xmin' in kwargs:
                        xlim[0] = kwargs['xmin']
                        limits_provided = True
                    
                    if 'xmax' in kwargs:
                        xlim[1] = kwargs['xmax']
                        limits_provided = True

                if 'ylim' in kwargs:
                    ylim = kwargs['ylim']
                    limits_provided = True
                else:
                    if 'ymin' in kwargs:
                        ylim[0] = kwargs['ymin']
                        limits_provided = True
                        
                    if 'ymax' in kwargs:
                        ylim[1] = kwargs['ymax']
                        limits_provided = True

                # If explicit limits are provided, use them to change the vlim ranges
                if limits_provided:
                    region_to_plot = np.zeros(self.dims).astype(bool)
                    region_to_plot[(xlim[0] <= X)*(X <= xlim[1])*(ylim[0] <= Y)*(Y <= ylim[1])] = True
                    vlim = [np.min(field[region_to_plot]), np.max(field[region_to_plot])]

                else:
                    vlim = [np.min(field), np.max(field)]
                
                # Set the value limitses
                if 'vlim' in kwargs:
                    vlim = kwargs['vlim']
                else:
                    if 'vmin' in kwargs:
                        vlim[0] = kwargs['vmin']
                    if 'vmax' in kwargs:
                        vlim[1] = kwargs['vmax']

                if vlim[1] - vlim[0] < 1e-10:
                    vlim = [vlim[0]-0.05, vlim[1]+0.05]

                pcm.set_clim(vmin=vlim[0], vmax=vlim[1])

                if 'vlim_symmetric' in kwargs:
                    vlim_symmetric = kwargs['vlim_symmetric']
                    if vlim_symmetric:
                        cmax = abs(field).max()
                        cmin = -cmax
                        pcm.set_clim(vmin=cmin, vmax=cmax)

                colorbar = kwargs.get('colorbar', True)

                if colorbar:
                    cbar = plt.colorbar(pcm, ax=ax)
            elif self.plot_lib == 'plotly':
                X, Y = np.meshgrid(self.x, self.y, indexing='ij')

                fig.add_trace(go.Heatmap(
                    x=X.flatten()/self.a0,
                    y=Y.flatten()/self.a0,
                    z=field.flatten(),
                    zmin=np.min(field),
                    zmax=np.max(field),
                    colorscale='Viridis', 
                    zsmooth='best',
                    hovertemplate='x: %{x:.2f} a₀<br>y: %{y:.2f} a₀<br> field: %{z:.2f}',
                    name=''
                ))

                if axis_equal:
                    # Set axis to be equal
                    fig.update_yaxes(
                        scaleanchor="x",
                        scaleratio=1,
                    )

                    fig.update_xaxes(
                        scaleanchor="y",
                        scaleratio=1,
                        ) 

        elif self.dim == 3:

            # Keyword arguments particular to the 3D case

            field_min = np.min(field)
            field_max = np.max(field)

            number_of_layers = kwargs.get('number_of_layers', 1)
            alpha = kwargs.get('alpha', 0.5)
            
            if 'vlim' in kwargs:
                vlim = kwargs['vlim']
                vmin = vlim[0]
                vmax = vlim[1]
            else:
                vmin = field_min
                vmax = field_max

            if 'layer_values' in kwargs:
                layer_values = np.concatenate([[-np.inf], kwargs['layer_values'], [np.inf]])
            else: 
                layer_values = np.linspace(vmin, vmax, number_of_layers + 2)

            if 'colormap' in kwargs:
                colormap = kwargs['colormap']
                if colormap == 'bluewhitered':
                    colormap = tool_colormap_bluewhitered()

                elif colormap == 'sunburst':
                    colormap = tool_colormap_sunburst()

                else:
                    colormap = plt.get_cmap(colormap)
            else: 
                colormap = plt.get_cmap('viridis')
            

            if 'colorbar' in kwargs:
                colorbar = kwargs['colorbar']
            else:
                colorbar = True

            #Plotting the layers
            if plot_lib == 'matplotlib':
                if ax == None:
                    plt.clf()
                    ax = plt.gcf().add_subplot(111, projection='3d')


                if field_min < layer_values[1] < field_max:
                    self.plot_tool_surface_matplotlib(field=field, 
                                            value=layer_values[1], 
                                            color=colormap((layer_values[1]-vmin) / (vmax-vmin)), 
                                            alpha=alpha,
                                            ax=ax)

                for layer_value in layer_values[2:-1]:
                    if field_min < layer_value < field_max:
                        self.plot_tool_surface_matplotlib(field=field, 
                                                value=layer_value, 
                                                color=colormap((layer_value-vmin) / (vmax-vmin)), 
                                                alpha=alpha,
                                                ax=ax)


                if colorbar:
                    sm = plt.cm.ScalarMappable(cmap=colormap)
                    sm.set_clim(vmin, vmax)
                    plt.colorbar(sm, ax=ax, pad=0.2)

            elif plot_lib == 'plotly':

                X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

                for layer_value in layer_values[1:-1]:

                    fig.add_trace(go.Isosurface(
                        x=X.flatten()/self.a0,
                        y=Y.flatten()/self.a0,
                        z=Z.flatten()/self.a0,
                        value = field.flatten(),
                        isomin = layer_value,
                        isomax = layer_value,
                        cmin = vmin,
                        cmax = vmax,
                        surface=dict(count=3),  # Ensuring only one surface is shown
                        colorscale='Viridis',
                        opacity=alpha,
                        showscale=bool(layer_value == layer_values[1])
                    )
                        )
        
        if plot_lib == 'matplotlib':
            kwargs['ax'] = ax
            self.plot_tool_set_axis_properties(**kwargs)
            return fig, ax

        elif plot_lib == 'plotly':
            kwargs['fig'] = fig
            self.plot_tool_set_axis_properties(**kwargs)
            return fig

        

    def plot_complex_field(self, complex_field, **kwargs):
        """
        Plot a complex field.

        ax=None, plot_method=None, colorbar=False

        Args:
            complex_field (numpy.ndarray): The complex field to plot.
            ax (matplotlib.axes.Axes, optional): The matplotlib axes on which to plot the field.
                If not provided, a new 3D axes will be created.
        
        Returns:
            matplotlib.axes.Axes: The axes containing the plot.
        """

        plot_lib = kwargs.get('plot_lib', self.plot_lib)

        if plot_lib == 'matplotlib':
            # Check if an axis object is provided
            fig = kwargs.get('fig', plt.gcf())
            ax = kwargs.get('ax', None)
        elif plot_lib == 'plotly':
            fig = kwargs.get('fig', go.Figure())

        # Kewyord arguments
        colorbar = kwargs.get('colorbar', True)
        axis_equal = kwargs.get('axis_equal',True)

        # Extend the field if not a complete array is given
        complex_field = self.plot_tool_extend_field(complex_field)

        # Calculate the magnitude and phase of the complex field
        rho = np.abs(complex_field)
        theta = np.angle(complex_field)

        if self.dim == 1:

            # Keyword arguments particular to the 1D case
            grid = kwargs.get('grid', False)

            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111)

                # Padding for the colorbar.
                padding=0.05

                ax.plot(self.x/self.a0, rho, color='black')

                # Color in the graph based on the argument of the complex field
                blend_factor=0.3 # The degree to which the color is blended with white
                cmap = tool_colormap_angle()

                ax.fill_between([self.xmin/self.a0,(self.xmin+self.dx/2)/self.a0], [rho[0],(rho[0]+rho[1])/2],
                                color=(1-blend_factor)*np.array(cmap((theta[0] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]), 
                                alpha=1)

                for i in range(1,self.xRes-1):
                    ax.fill_between([(self.x[i]-self.dx/2)/self.a0,self.x[i]/self.a0], [(rho[i]+rho[i-1])/2,rho[i]],
                                    color=(1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]), 
                                    alpha=1)
                    ax.fill_between([self.x[i]/self.a0,(self.x[i]+self.dx/2)/self.a0], [rho[i],(rho[i]+rho[i+1])/2],
                        color=(1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]),  
                        alpha=1)

                ax.fill_between([(self.xmax-1.5*self.dx)/self.a0,(self.xmax-self.dx)/self.a0], [(rho[-1]+rho[-2])/2,rho[-1]],
                                color=(1-blend_factor)*np.array(cmap((theta[-1] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1]),  
                                alpha=1)

            elif self.plot_lib == 'plotly':
                
                if fig == None:
                    fig = go.Figure()

                # Color in the graph based on the argument of the complex field
                blend_factor=0.3 # The degree to which the color is blended with white
                cmap = tool_colormap_angle()

                color = (1-blend_factor)*np.array(cmap((theta[0] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
                color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')

                fig.add_trace(go.Scatter(x=[self.xmin/self.a0,(self.xmin+self.dx/2)/self.a0], 
                              y=[rho[0],(rho[0]+rho[1])/2],
                                mode='lines',
                                line=dict(color='rgba(0,0,0,0)'),
                                fill='tozeroy',
                                showlegend=False,
                                hoverinfo='skip',
                                fillcolor=color_str))

                for i in range(1,self.xRes-1):
                    color = (1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
                    color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
                    fig.add_trace(go.Scatter(x=[(self.x[i]-self.dx/2)/self.a0,self.x[i]/self.a0], 
                              y=[(rho[i]+rho[i-1])/2,rho[i]],
                                mode='lines',
                                line=dict(color='rgba(0,0,0,0)'),
                                fill='tozeroy',
                                showlegend=False,
                                hoverinfo='skip',
                                fillcolor=color_str))

                    color = (1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
                    color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
                    fig.add_trace(go.Scatter(x=[self.x[i]/self.a0,(self.x[i]+self.dx/2)/self.a0], 
                              y=[rho[i],(rho[i]+rho[i+1])/2],
                                mode='lines',
                                line=dict(color='rgba(0,0,0,0)'),
                                fill='tozeroy',
                                showlegend=False,
                                hoverinfo='skip',
                                fillcolor=color_str))

                color = (1-blend_factor)*np.array(cmap((theta[-1] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
                color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
                fig.add_trace(go.Scatter(x=[(self.xmax-1.5*self.dx)/self.a0,(self.xmax-self.dx)/self.a0], 
                              y=[(rho[-1]+rho[-2])/2,rho[-1]],
                                mode='lines',
                                line=dict(color='rgba(0,0,0,0)'),
                                fill='tozeroy',
                                showlegend=False,
                                hoverinfo='skip',
                                fillcolor=color_str))

                fig.add_trace(go.Scatter(
                    x=self.x/self.a0,
                    y=rho,
                    mode='lines',
                    showlegend=False,
                    customdata=np.stack((theta/np.pi, rho), axis=-1),
                    hovertemplate='x: %{x:.2f} a₀<br>θ: %{customdata[0]:.2f} π<br>ρ: %{customdata[1]:.2e}',
                    name='',
                    line=dict(color='black')
                ))
                

        elif self.dim == 2:
            # Keyword arguments particular to the 2D case
            plot_method = kwargs.get('plot_method', 'phase_angle')

            # Create a meshgrid
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')

            if plot_method == '3Dsurface':
                if self.plot_lib == 'matplotlib':  
                    # Padding for the colorbar
                    padding=0.2

                    # Keyword arguments particular to the 3D surface plot
                    grid = kwargs.get('grid', True)
                    kwargs['axis_equal'] = False
                    
                    if ax == None:
                        fig.clf()
                        ax = fig.add_subplot(111, projection='3d')
                    
                    # Get the colors from a colormap (e.g., hsv, but you can choose any other)
                    colors = tool_colormap_angle()((theta + np.pi) / (2 * np.pi))  # Normalizing theta to [0, 1]

                    surf = ax.plot_surface(X/self.a0, Y/self.a0, rho, facecolors=colors)

                elif self.plot_lib == 'plotly':
                    print('Plotly not yet implemented for 2D 3Dsurface angle plots.')
                    pass

            elif plot_method == 'phase_angle':
                

                # Keyword arguments particular to the phase angle plot
                grid = kwargs.get('grid', False)

                rho_normalized = rho / np.max(rho)

                custom_colormap = tool_colormap_angle()
                
                if self.plot_lib == 'matplotlib':
                    
                    # Check if an axis object is provided
                    if ax == None:
                        fig.clf()
                        ax = fig.add_subplot(111)

                    # Padding for the colorbar
                    padding=0.05
                    
                    mesh = ax.pcolormesh(X/self.a0, Y/self.a0, theta, shading='auto', cmap=custom_colormap, vmin=-np.pi, vmax=np.pi)
                    mesh.set_alpha(rho_normalized)
                
                elif self.plot_lib == 'plotly':

                    norm = mcolors.Normalize(vmin=-np.pi, vmax=np.pi)
                    colormap = cm.ScalarMappable(norm=norm, cmap=custom_colormap)

                    # Create an RGBA image array
                    rgba_image = colormap.to_rgba(theta)

                    # Apply the spatial opacity
                    for i in range(rgba_image.shape[0]):
                        for j in range(rgba_image.shape[1]):
                            for k in range(3):
                                rgba_image[i, j, k] = 1-(1-rgba_image[i, j, k])*rho_normalized[i, j]  # Set alpha channel

                    rgba_image = np.transpose(rgba_image, (1, 0, 2))

                    # Convert RGBA image to a format Plotly can understand
                    image_data = (rgba_image * 255).astype(np.uint8)


                    if fig == None:
                        fig = go.Figure()

                    fig.add_trace(go.Image(z=image_data, 
                                    dx=self.dx/self.a0, 
                                    dy=self.dy/self.a0, 
                                    x0=self.xmin/self.a0, 
                                    y0=self.ymin/self.a0,
                                    hovertemplate='x: %{x:.2f} a₀<br>y: %{y:.2f} a₀<br>θ: %{customdata[0]:.2f} π<br>ρ: %{customdata[1]:.2e}',
                                    customdata=np.stack((np.transpose(theta/np.pi), np.transpose(rho)), axis=-1),
                                    name=''
                                    ))  


                    # fig.add_trace(go.Heatmap(
                    #     x=X.flatten()/self.a0,
                    #     y=Y.flatten()/self.a0,
                    #     z=theta.flatten(),
                    #     zmin=-np.pi,
                    #     zmax=np.pi,
                    #     colorscale=tool_colormap_angle(pyplot=True),
                    #     colorbar=dict(
                    #             tickvals=[-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi],
                    #             ticktext=['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
                    #         )
                    # ))

                    if axis_equal:
                       # Set axis to be equal
                        fig.update_yaxes(
                            scaleanchor="x",
                            scaleratio=1,
                        )

                        fig.update_xaxes(
                            scaleanchor="y",
                            scaleratio=1,
                            ) 

                    

                    

        elif self.dim == 3:

            grid = kwargs.get('grid', True)

            plot_method = kwargs.get('plot_method', 'phase_blob')
            
            X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

            rho_normalized = rho / np.max(rho)

            colormap = tool_colormap_angle()

            if plot_lib == 'matplotlib':
                ax = kwargs.get('ax', None)

                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')
            
            elif plot_lib == 'plotly':
                pass
        
            if plot_method == 'phase_angle':
                
                if self.plot_lib == 'matplotlib':
                    
                    # Padding for the colorbar
                    padding=0.2

                    for angle in [-2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3]:
                        field_to_plot = theta.copy()
                        field_to_plot[theta < angle - 1] = float('nan')
                        field_to_plot[theta > angle + 1] = float('nan')
                        field_to_plot[rho_normalized < 0.01] = float('nan')

                        if np.nanmin(field_to_plot) < angle < np.nanmax(field_to_plot):
                            #TODO: make alpha a keyword argument (Vidar 11.03.24)
                            self.plot_tool_surface_matplotlib(field=field_to_plot, 
                                                value=angle, 
                                                color=colormap((angle + np.pi) / (2 * np.pi)), 
                                                alpha=0.5,
                                                ax=ax)

                    theta = np.mod(theta, 2 * np.pi)

                    field_to_plot = theta.copy()
                    field_to_plot[theta < np.pi - 1] = float('nan')
                    field_to_plot[theta > np.pi + 1] = float('nan')
                    field_to_plot[rho_normalized < 0.01] = float('nan')

                    if np.nanmin(field_to_plot) < np.pi < np.nanmax(field_to_plot):
                        #TODO: make alpha a keyword argument (Vidar 11.03.24)
                        self.plot_tool_surface_matplotlib(field=field_to_plot, 
                                                value=np.pi, 
                                                color=colormap(0), 
                                                alpha=0.5,
                                                ax=ax)
                elif self.plot_lib == 'plotly':
                    print('Plotly not yet implemented for 3D phase angle plots.')
                    pass
            
            elif plot_method == 'phase_blob':
                # TODO: change the function so that it used plot_tool_surface_matplotlib (Vidar 11.03.24)
                # Padding for the colorbar
                padding=0.2
                
                phase_blob_threshold = kwargs.get('phase_blob_threshold', 0.5)

                if np.nanmin(rho_normalized)<phase_blob_threshold<np.nanmax(rho_normalized):
                    verts, faces, _, _ = marching_cubes(rho_normalized, phase_blob_threshold)

                    # Calculate the centroids of each triangle
                    centroids = np.mean(verts[faces], axis=1)

                    # Assuming theta is defined on the same grid as rho
                    x, y, z = np.mgrid[0:rho_normalized.shape[0], 0:rho_normalized.shape[1], 0:rho_normalized.shape[2]]

                    # Flatten the grid for interpolation
                    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
                    theta_values = theta.ravel()

                    # Interpolate theta at the vertices positions
                    theta_faces = sp.interpolate.griddata(points, theta_values, centroids, method='nearest')

                    # Normalize theta values for color mapping
                    theta_faces_normalized = (theta_faces + np.pi) / (2*np.pi)

                    # Map normalized theta values to colors
                    colors = colormap(theta_faces_normalized)

                    if plot_lib == 'matplotlib':
                        

                        # Plot the complex field
                        ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0, 
                                        (self.ymin+verts[:, 1]*self.dy)/self.a0, 
                                        faces, 
                                        (self.zmin+verts[:, 2]*self.dz)/self.a0, 
                                        facecolor=colors, antialiased=False)
                    
                        # Plot the shadows on the edges
                        plot_shadows = kwargs.get('plot_shadows', True)
                        if plot_shadows:
                            ax.plot_trisurf((self.xmin+0*verts[:, 0]*self.dx)/self.a0, 
                                            (self.ymin+verts[:, 1]*self.dy)/self.a0, 
                                            faces, 
                                            (self.zmin+verts[:, 2]*self.dz)/self.a0, 
                                            facecolor='black', antialiased=True,
                                            alpha=0.1)

                            ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0, 
                                            (self.ymax+0*verts[:, 1]*self.dy)/self.a0, 
                                            faces, 
                                            (self.zmin+verts[:, 2]*self.dz)/self.a0, 
                                            facecolor='black', antialiased=True,
                                            alpha=0.1)
                            
                            ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0, 
                                            (self.ymin+verts[:, 1]*self.dy)/self.a0, 
                                            faces, 
                                            (self.zmin+0*verts[:, 2]*self.dz)/self.a0, 
                                            facecolor='black', antialiased=True,
                                            alpha=0.1)

                    elif plot_lib == 'plotly':
                        # Convert colors to 'rgba()' format required by Plotly
                        colors = ['rgba({},{},{},{})'.format(*c) for c in (255*colors).astype(int)]

                        # Create the mesh object
                        x_new = (verts[:, 0] * self.dx + self.xmin) / self.a0
                        y_new = (verts[:, 1] * self.dy + self.ymin) / self.a0
                        z_new = (verts[:, 2] * self.dz + self.zmin) / self.a0

                        mesh = go.Mesh3d(
                            x=x_new, 
                            y=y_new, 
                            z=z_new,
                            i=faces[:, 0], 
                            j=faces[:, 1], 
                            k=faces[:, 2],
                            facecolor=colors,  # Set color for each face
                            showscale=True
                        )

                        fig = go.Figure(data=[mesh])

        if colorbar:
            if plot_lib == 'matplotlib':
                mappable = plt.cm.ScalarMappable(cmap=tool_colormap_angle())
                mappable.set_array([])
                mappable.set_clim(-np.pi, np.pi)
                cbar = plt.colorbar(mappable, ax=ax, pad=padding)
                cbar.set_ticks(np.array([-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]))
                cbar.set_ticklabels([r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$'])

            elif plot_lib == 'plotly':
                # Add a colorbar to the plot
                if self.dim == 1:
                    pass
                elif self.dim == 2:
                    pass
                elif self.dim == 3:
                    fig = self.plot_tool_plotly_add_angle_colorbar3D(fig)

        if plot_lib == 'matplotlib':
            kwargs['ax'] = ax
            kwargs['grid'] = grid
            self.plot_tool_set_axis_properties(**kwargs)
            return fig, ax

        elif plot_lib == 'plotly':
            kwargs['fig'] = fig
            self.plot_tool_set_axis_properties(**kwargs)
            return fig

    def plot_angle_field(self, angle_field, **kwargs):
        """Plot the angle field.

        Args:
            field (array-like): The angle field values.
            ax (matplotlib.axes.Axes, optional): The axes to plot the angle field on. If not provided, a new subplot will be created.
        
        Returns:
            The axes containing the plot. (matplotlib.axes.Axes)
        """

        # Check if the vector field is complex
        if np.iscomplexobj(angle_field):
            print("\033[91mWarning: the angle vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
            print('Max imaginary part: ', np.max(np.imag(angle_field)))
            angle_field = np.real(angle_field)

        # Extend the field if not a complete array is given
        angle_field = self.plot_tool_extend_field(angle_field)

        # Normalize around 0
        angle_field = np.mod(angle_field, 2 * np.pi) - np.pi        

        if self.dim == 1:
            if 'vlim' in kwargs:
                vlim = kwargs['vlim']
            else:
                kwargs['vlim'] = [-np.pi, np.pi]
                kwargs['yticks'] = [-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]
                kwargs['yticklabels'] = [r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$']

            
            
            return self.plot_field(angle_field, **kwargs)
        
        elif self.dim > 1:
            complex_field = np.exp(1j * angle_field)

            kwargs['plot_method'] = 'phase_angle'

            return self.plot_complex_field(complex_field, **kwargs)

    def plot_vector_field(self, vector_field, **kwargs):
        """Plots a vector field.

        Args:
            vector_field (tuple): Tuple containing the x and y components of the vector field.
            **kwargs: 
                spacing (int, optional): The spacing for the quiver plot. Default is self.xRes//20.
                Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

        Returns:
            Tuple containing
                - The figure containing the plot.
                - The axes containing the plot.
        """
        spacing = kwargs.get('spacing', self.xRes//20)

        # Extend the field if not a complete array is given
        vector_field_copy = []
        for n in range(len(vector_field)):
            print(vector_field[n].shape)
            vector_field_copy.append(self.plot_tool_extend_field(vector_field[n]))
        vector_field = np.array(vector_field_copy)
        
        # Keyword arguments
        axis_equal = kwargs.get('axis_equal',True)
        colormap = kwargs.get('colormap', 'viridis')

        # Check if the vector field is complex
        if np.iscomplexobj(vector_field):
            print("\033[91mWarning: the provided vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
            print('Max imaginary part: ', np.max(np.imag(vector_field)))
            vector_field = np.real(vector_field)

        if self.plot_lib == 'matplotlib':
            # Check if an axis object is provided
            fig = kwargs.get('fig', plt.gcf())
            ax = kwargs.get('ax', None)

        def add_spacing_2D(X,Y,U,V,spacing):
            X = X[::spacing, ::spacing]
            Y = Y[::spacing, ::spacing]
            U = U[::spacing, ::spacing]
            V = V[::spacing, ::spacing]
            return X,Y,U,V

        def add_spacing_3D(X,Y,Z,U,V,W,spacing):
            X = X[::spacing, ::spacing, ::spacing]
            Y = Y[::spacing, ::spacing, ::spacing]
            Z = Z[::spacing, ::spacing, ::spacing]
            U = U[::spacing, ::spacing, ::spacing]
            V = V[::spacing, ::spacing, ::spacing]
            W = W[::spacing, ::spacing, ::spacing]
            return X,Y,Z,U,V,W            


        ###############################################################
        ########### DIMENSION: 1 - VECTOR-DIMENSION: 1 ################
        ###############################################################
            
        if self.dim == 1 and vector_field.shape == (1,self.xRes):

            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = plt.gcf().add_subplot(111)

            X, Y = np.meshgrid(self.x, np.array([0]), indexing='ij')

            U = np.zeros(X.shape)
            V = np.zeros(X.shape)

            V[:,0] = vector_field[0]

            X,Y,U,V = add_spacing_2D(X,Y,U,V,spacing)

            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, U, V, color='blue', angles='xy', scale_units='xy', scale=1)
            
            kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]

        ###############################################################
        ########### DIMENSION: 1 - VECTOR-DIMENSION: 2 ################
        ###############################################################

        elif self.dim == 1 and vector_field.shape == (2,self.xRes):
            
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')
            
            X, Y, Z = np.meshgrid(self.x, np.array([0]), np.array([0]), indexing='ij')

            U = np.zeros(X.shape)
            V = np.zeros(X.shape)
            W = np.zeros(X.shape)

            V[:,0,0] = vector_field[0]
            W[:,0,0] = vector_field[1]

            X,Y,Z,U,V,W = add_spacing_3D(X,Y,Z,U,V,W,spacing)

            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')

            kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]
            delta_y = kwargs['ylim'][1] - kwargs['ylim'][0]

            kwargs['zlim'] = [np.min(vector_field[1]), np.max(vector_field[1])]
            delta_z = kwargs['zlim'][1] - kwargs['zlim'][0]

            if delta_y < 0.15*delta_z:
                kwargs['ylim'] = [kwargs['ylim'][0] - 0.15*delta_z, kwargs['ylim'][1] + 0.15*delta_z]
            
            if delta_z < 0.15*delta_y:
                kwargs['zlim'] = [kwargs['zlim'][0] - 0.15*delta_y, kwargs['zlim'][1] + 0.15*delta_y]

        ###############################################################
        ########### DIMENSION: 1 - VECTOR-DIMENSION: 3 ################
        ###############################################################

        elif self.dim == 1 and vector_field.shape == (3,self.xRes):
            
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')
            
            X, Y, Z = np.meshgrid(self.x, np.array([0]), np.array([0]), indexing='ij')

            U = np.zeros(X.shape)
            V = np.zeros(X.shape)
            W = np.zeros(X.shape)

            # Add the vector field
            U[:,0,0] = vector_field[0]
            V[:,0,0] = vector_field[1]
            W[:,0,0] = vector_field[2]

            # Add spacing
            X,Y,Z,U,V,W = add_spacing_3D(X,Y,Z,U,V,W,spacing)

            # Normalize the vectors
            max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

            # Normalizing
            U = U/max_vector
            V = V/max_vector
            W = W/max_vector

            # Scale factors
            vx_scale = kwargs.get('vx_scale', spacing*self.xmax/15/self.a0)
            vy_scale = kwargs.get('vy_scale', 1)
            vz_scale = kwargs.get('vz_scale', 1)

            # Scaling
            U = vx_scale*U
            V = vy_scale*V
            W = vz_scale*W

            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
            
            kwargs['ylim'] = [-1,1]
            kwargs['zlim'] = [-1,1]

        ###############################################################
        ########### DIMENSION: 2 - VECTOR-DIMENSION: 1 ################
        ###############################################################

        elif self.dim == 2 and vector_field.shape == (1,self.xRes,self.yRes):
            
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = plt.gcf().add_subplot(111)

            X, Y = np.meshgrid(self.x, self.y, indexing='ij')

            U = vector_field[0]
            V = np.zeros(X.shape)

            X,Y,U,V = add_spacing_2D(X,Y,U,V,spacing)

            # Normalize the vectors
            max_vector = np.max(np.sqrt(U ** 2))
            
            U = U / max_vector

            # Scale factors
            vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)

            # Scaling
            U = vx_scale*U
            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, U, V, color='blue',
                    angles='xy', scale_units='xy', scale=1,
                    headwidth=3, headlength=4, headaxislength=3, pivot='middle')

        ###############################################################
        ########### DIMENSION: 2 - VECTOR-DIMENSION: 2 ################
        ###############################################################

        elif self.dim == 2 and vector_field.shape == (2,self.xRes,self.yRes):
            
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = plt.gcf().add_subplot(111)

            X, Y = np.meshgrid(self.x, self.y, indexing='ij')

            X, Y, U, V = add_spacing_2D(X,Y,vector_field[0],vector_field[1],spacing)

            if self.plot_lib == 'matplotlib':

                max_vector = np.max(np.sqrt(U ** 2 + V ** 2))
            
                U = U / max_vector
                V = V / max_vector

                # Scale factors
                vx_scale = kwargs.get('vx_scale', 1.0*spacing*self.dx/self.a0)
                vy_scale = kwargs.get('vy_scale', 1.0*spacing*self.dy/self.a0)

                # Scaling
                U = vx_scale*U
                V = vy_scale*V

                ax.quiver(X/self.a0, Y/self.a0, U, V, color='blue', 
                    angles='xy', scale_units='xy', scale=1,
                    headwidth=3, headlength=4, headaxislength=3, pivot='middle')

            
            elif self.plot_lib == 'plotly':
                
                # An attempt to use Cone. Not optimal
                # fig = kwargs.get('fig', go.Figure())
                # fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                #             y=Y.flatten()/self.a0, 
                #             z=np.zeros(X.shape).flatten(), 
                #             u=U.flatten(), 
                #             v=V.flatten(), 
                #             w=np.zeros(X.shape).flatten(),
                #             colorscale='Viridis', 
                #             sizemode='scaled', 
                #             sizeref=1, showscale=True))
                # fig.update_layout(
                #         scene_camera=dict(
                #             eye=dict(x=0, y=0, z=2.5),
                #             up=dict(x=0, y=0, z=1)
                #         ),
                #         scene=dict(
                #             zaxis=dict(showticklabels=False, title='')
                #         )
                #     )

                def get_colors(values, colorscale='Viridis'):
                    colorscale = pc.get_colorscale(colorscale)
                    unique_magnitudes = np.unique(values)
                    color_map = {val: pc.sample_colorscale(colorscale, val)[0] for val in unique_magnitudes}
                    return np.vectorize(color_map.get)(values)

                    
                def plot_triangle(fig, position,direction,size,color):
                    x = [position[0]+direction[0]*size/2, 
                            position[0]-direction[0]*size/3 + direction[1]*size/4, 
                            position[0]-direction[0]*size/3 - direction[1]*size/4]
                    y = [position[1]+direction[1]*size/2, 
                            position[1]-direction[1]*size/3 - direction[0]*size/4, 
                            position[1]-direction[1]*size/3 + direction[0]*size/4]

                    fig.add_trace(go.Scatter(
                        x=x,
                        y=y,
                        fill='toself',
                        mode='lines', 
                        line=dict(color='rgba(0,0,0,0)'),
                        fillcolor=color,
                        showlegend=False,
                        name=''
                    ))

                fig = kwargs.get('fig', go.Figure())

                u = U.flatten()
                v = V.flatten()
                
                magnitude = np.sqrt(u**2 + v**2)
                magnitude_normalized = magnitude/np.max(magnitude)

                colors = get_colors(magnitude_normalized, colorscale='viridis')

                angle = np.arctan2(v, u)
                
                direction = np.array([np.cos(angle), np.sin(angle)]).T

                x = X.flatten()/self.a0
                y = Y.flatten()/self.a0

                for i in range(len(x)):
                    plot_triangle(fig, position=[x[i],y[i]], direction=direction[i], size=1.3*magnitude_normalized[i]*spacing*self.dx/self.a0, color=colors[i])

                colorbar = kwargs.get('colorbar', True)

                if colorbar:
                    # Add a scatter trace to create the colorbar
                    colorbar_trace = go.Scatter(
                        x=[None],
                        y=[None],
                        mode='markers',
                        marker=dict(
                            colorscale=colormap,
                            cmin=0,
                            cmax=np.max(magnitude),
                            colorbar=dict(
                                title=''
                            )
                        ),
                        showlegend=False
                    )
                    fig.add_trace(colorbar_trace)

        ###############################################################
        ########### DIMENSION: 2 - VECTOR-DIMENSION: 3 ################
        ###############################################################

        elif self.dim == 2 and vector_field.shape == (3,self.xRes,self.yRes):

            if self.plot_lib == 'plotly':
                print('Plotly not yet implemented for this type of vector field.')
                pass

            elif self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')

            X, Y, Z = np.meshgrid(self.x, self.y, np.array([0]), indexing='ij')
            U = np.zeros(X.shape)
            V = np.zeros(X.shape)
            W = np.zeros(X.shape)
            
            U[:,:,0] = vector_field[0]
            V[:,:,0] = vector_field[1]
            W[:,:,0] = vector_field[2]

            X,Y,Z,U,V,W = add_spacing_3D(X,Y,Z,U,V,W,spacing)

            max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

            # Normalizing
            U = U / max_vector
            V = V / max_vector
            W = W / max_vector
            
            # Scale factors
            vx_scale = kwargs.get('vx_scale', 2*spacing*self.xmax/max_vector/self.a0)
            vy_scale = kwargs.get('vy_scale', 2*spacing*self.ymax/max_vector/self.a0)
            vz_scale = kwargs.get('vz_scale', spacing/self.a0)

            # Scaling
            U = vx_scale*U
            V = vy_scale*V
            W = vz_scale*W
            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')

            kwargs['axis_equal'] = False
            kwargs['zlim'] = [-spacing,spacing]

        ###############################################################
        ########### DIMENSION: 3 - VECTOR-DIMENSION: 1 ################
        ###############################################################

        elif self.dim == 3 and vector_field.shape == (1,self.xRes,self.yRes,self.zRes):

            X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')

            # Define the vector field
            U = vector_field[0]
            V = np.zeros(U.shape)
            W = np.zeros(U.shape)

            # Add spacing
            X,Y,Z,U,V,W = add_spacing_3D(X,Y,Z,U,V,W,spacing)

            # Normalize the vectors
            max_vector = np.max(np.sqrt(U ** 2))
            U = U / max_vector

            # Scale factors
            vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)

            # Scaling
            U = vx_scale*U

            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
            elif self.plot_lib == 'plotly':
                print('Plotly not yet implemented for this type of vector field.')
                pass

        ###############################################################
        ########### DIMENSION: 3 - VECTOR-DIMENSION: 2 ################
        ###############################################################

        elif self.dim == 3 and vector_field.shape == (2,self.xRes,self.yRes,self.zRes):

            X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')
            
            # Define the vector field
            U = vector_field[0]
            V = vector_field[1]
            W = np.zeros(U.shape)

            # Add spacing
            X,Y,Z,U,V,W = add_spacing_3D(X,Y,Z,U,V,W,spacing)

            # Normalize the vectors
            max_vector = np.max(np.sqrt(U ** 2 + V ** 2))
            U = U / max_vector
            V = V / max_vector

            # Scale factors
            vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)
            vy_scale = kwargs.get('vy_scale', spacing*self.dy/self.a0)

            # Scaling
            U = vx_scale*U
            V = vy_scale*V

            if self.plot_lib == 'matplotlib':
                ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
            elif self.plot_lib == 'plotly':
                print('Plotly not yet implemented for this type of vector field.')
                pass

        ###############################################################
        ########### DIMENSION: 3 - VECTOR-DIMENSION: 3 ################
        ###############################################################

        elif self.dim == 3 and vector_field.shape == (3,self.xRes,self.yRes,self.zRes):

            X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')
            if self.plot_lib == 'matplotlib':
                if ax == None:
                    fig.clf()
                    ax = fig.add_subplot(111, projection='3d')

                    
            
            # Define the vector field
            U = vector_field[0]
            V = vector_field[1]
            W = vector_field[2]

            X,Y,Z,U,V,W = add_spacing_3D(X,Y,Z,U,V,W,spacing)

            

            if self.plot_lib == 'matplotlib':
                # Normalize the vectors
                max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

                U = U / max_vector
                V = V / max_vector
                W = W / max_vector

                # Scale factors
                vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)
                vy_scale = kwargs.get('vy_scale', spacing*self.dy/self.a0)
                vz_scale = kwargs.get('vz_scale', spacing*self.dz/self.a0)

                # Scaling
                U = vx_scale*U
                V = vy_scale*V
                W = vz_scale*W
                ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
            
            elif self.plot_lib == 'plotly':
                fig = kwargs.get('fig', go.Figure())
                fig.add_trace(go.Cone(x=X.flatten()/self.a0, y=Y.flatten()/self.a0, z=Z.flatten()/self.a0, u=U.flatten(), v=V.flatten(), w=W.flatten(), colorscale='Viridis', sizemode='scaled', sizeref=1, showscale=True))


        ###############################################################
        ###########     NON-VALID DIMENSION            ################
        ###############################################################

        else:
            raise Exception("You have entered an invalid field to the plot_vector_field function.")

        if self.plot_lib == 'matplotlib':
            kwargs['ax'] = ax
            self.plot_tool_set_axis_properties(**kwargs)
            return fig, ax

        elif self.plot_lib == 'plotly':
            kwargs['fig'] = fig
            self.plot_tool_set_axis_properties(**kwargs)
            return fig

    def plot_field_in_plane(self, field, normal_vector=None, position=None, 
                        **kwargs):
        """Plots the field in a plane perpendicular to the given normal vector
        
        Uses scipy.interpolate.griddata and plt.plot_trisurf.

        Args:
            field (array-like): The field to be plotted.
            normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
            position (array-like, optional): The position of the plane. Default is the middle of the system.
            **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/. 
        
        Returns:
            The axes containing the plot. (matplotlib.axes.Axes)
        """

        if self.dim != 3:
            raise Exception("The plot in plane function is only defined for 3D fields.")

        # Extend the field if not a complete array is given
        field = self.plot_tool_extend_field(field)

        # Check if the vector field is complex
        if np.iscomplexobj(field):
            print("\033[91mWarning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
            print('Max imaginary part: ', np.max(np.imag(field)))
            field = np.real(field)

        # Check if an axis object is provided
        fig = kwargs.get('fig', plt.gcf())
        ax = kwargs.get('ax', None)

        # Kewyord arguments
        colorbar = kwargs.get('colorbar', True)

        # Default values of position and normal vector
        if position is None:
            position = self.rmid

        if normal_vector is None:
            normal_vector=[0,1,0]

        if ax is None:
            ax = fig.add_subplot(111, projection='3d')

        colormap = kwargs.get('colormap', 'viridis')

        if colormap == 'angle':
            colormap = tool_colormap_angle()
        elif colormap == 'bluewhitered':
            colormap = tool_colormap_bluewhitered()
        else:
            colormap = plt.get_cmap(colormap)

        normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
        height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

        verts, faces, _, _ = marching_cubes(height_above_plane, 0)

        # Calculate the centroids of each triangle
        centroids = np.mean(verts[faces], axis=1)

        # Assuming field is defined on the same grid as height_above_plane
        x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

        # Flatten the grid for interpolation
        points = np.c_[x.ravel(), y.ravel(), z.ravel()]
        field_values = field.ravel()

        # Interpolate field at the vertices positions
        field_verts = sp.interpolate.griddata(points, field_values, centroids, method='nearest')

        # Normalize field values for color mapping
        field_normalized = (field_verts - np.min(field_verts)) / (np.max(field_verts) - np.min(field_verts))

        # Map normalized field values to colors
        colors = colormap(field_normalized)
    
        ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0,
                        (self.ymin+verts[:, 1]*self.dy)/self.a0,
                        faces,
                        (self.zmin+verts[:, 2]*self.dz)/self.a0,
                        facecolor=colors, antialiased=False)

        if colorbar:
            sm = plt.cm.ScalarMappable(cmap=colormap)  
            sm.set_clim(np.min(field_verts),np.max(field_verts))  
            cbar = plt.colorbar(sm, ax=ax, pad=0.2)

        kwargs['grid'] = kwargs.get('grid', True)
        kwargs['ax'] = ax
        self.plot_tool_set_axis_properties(**kwargs)

        return fig, ax

    def plot_complex_field_in_plane(self, complex_field, normal_vector=None, position=None, **kwargs):
        """Plots the complex field in a plane perpendicular to the given normal vector using

        Args:
            complex_field (array-like): The complex field to be plotted.
            normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
            position (array-like, optional): The position of the plane. Default is the middle of the system.
            **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

        Returns:
            The axes containing the plot of the complex field. (matplotlib.axes.Axes)
        """

        if self.dim != 3:
            raise Exception("The plot in plane function is only defined for 3D fields.")

        # Extend the field if not a complete array is given
        complex_field = self.plot_tool_extend_field(complex_field)

        # Default values of position and normal vector
        if position is None:
            position = self.rmid

        if normal_vector is None:
            normal_vector = [0,1,0]

        # Calculate the magnitude and phase of the complex field
        rho = np.abs(complex_field)
        theta = np.angle(complex_field)
        
        # Check if an axis object is provided
        fig = kwargs.get('fig', plt.gcf())
        ax = kwargs.get('ax', None)

        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')

        # Kewyord arguments
        colorbar = kwargs.get('colorbar', True)

        normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
        height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

        verts, faces, _, _ = marching_cubes(height_above_plane, 0)

        # Calculate the centroids of each triangle
        centroids = np.mean(verts[faces], axis=1)

        # Assuming field is defined on the same grid as height_above_plane
        x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

        # Flatten the grid for interpolation
        points = np.c_[x.ravel(), y.ravel(), z.ravel()]
        theta_values = theta.ravel()

        # Interpolate field at the vertices positions
        theta_verts = sp.interpolate.griddata(points, theta_values, centroids, method='nearest')
        rho_verts = sp.interpolate.griddata(points, rho.ravel(), centroids, method='nearest')

        # Normalize field values for color mapping
        theta_normalized = (theta_verts+np.pi) / (2*np.pi)

        # Map normalized field values to colors
        colormap = tool_colormap_angle()
        colors = colormap(theta_normalized)

        # Blend the colors with white according to rho (normalized)
        colors[:,3] = (rho_verts/np.max(rho_verts)).ravel()
    
        ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0,
                        (self.ymin+verts[:, 1]*self.dy)/self.a0,
                        faces,
                        (self.zmin+verts[:, 2]*self.dz)/self.a0,
                        facecolor=colors, antialiased=True)

        # Create a colorbar
        if colorbar:
            mappable = plt.cm.ScalarMappable(cmap=tool_colormap_angle())
            mappable.set_array([])
            mappable.set_clim(-np.pi, np.pi)
            cbar = plt.colorbar(mappable, ax=ax, pad=0.2)
            cbar.set_ticks(np.array([-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]))
            cbar.set_ticklabels([r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$'])

        kwargs['grid'] = kwargs.get('grid', True)
        kwargs['ax'] = ax
        self.plot_tool_set_axis_properties(**kwargs)

        return fig, ax

    def plot_angle_field_in_plane(self, angle_field, normal_vector=None, position=None,**kwargs):
        """Plots the angle field in a plane.

        Args:
            angle_field (numpy.ndarray): The angle field to be plotted.
            normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
            position (array-like, optional): The position of the plane. Default is the middle of the system.
            **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

        Returns:
            The axes containing the plot. (matplotlib.axes.Axes)
        """

        # Check if the vector field is complex
        if np.iscomplexobj(angle_field):
            print("\033[91mWarning: the angle vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
            print('Max imaginary part: ', np.max(np.imag(angle_field)))
            angle_field = np.real(angle_field)

        # Extend the field if not a complete array is given
        angle_field = self.plot_tool_extend_field(angle_field)

        complex_field = np.exp(1j * angle_field)
    
        return self.plot_complex_field_in_plane(complex_field, normal_vector=normal_vector, position=position, **kwargs)

    
    def plot_vector_field_in_plane(self,vector_field,position=None,normal_vector=None,spacing=2,**kwargs):
        """
        Plots the vector field in a plane.
        
        Args:
            vector_field (tuple): Tuple containing the x and y components of the vector field.
            position (array-like, optional): The position of the plane. Default is the middle of the system.
            normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
            spacing (int, optional): The spacing for the quiver plot. Default is 5.
            kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

        Returns:
            Tuple consisting of
                - The figure containing the plot. matplotlib.figure.Figure.
                - The axes containing the plot. matplotlib.axes.Axes: 
        """

        # Convert the vector field to a numpy array
        vector_field = np.array(vector_field)

        # Extend the field if not a complete array is given
        # TODO: Make this part more efficient (Vidar 11.03.24)
        vector_field_copy = []
        for n in range(vector_field.shape[0]):
            vector_field_copy.append(self.plot_tool_extend_field(vector_field[n]))
        vector_field = np.array(vector_field_copy)

        # Check if the vector field is complex
        if np.iscomplexobj(vector_field):
            print("\033[91mWarning: the provided vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
            print('Max imaginary part: ', np.max(np.imag(vector_field)))
            vector_field = np.real(vector_field)
        

        if self.dim != 3:
            raise Exception("The plot in plane function is only defined for 3D fields.")

        # Check if an axis object is provided
        fig = kwargs.get('fig', plt.gcf())
        ax = kwargs.get('ax', None)

        if ax == None:
            ax = fig.add_subplot(111, projection='3d')

        # Default values of position and normal vector
        if position is None:
            position = self.rmid
        
        if normal_vector is None:
            normal_vector = [0,1,0]
        
        normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
        height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

        verts, faces, _, _ = marching_cubes(height_above_plane, 0)

        # Assuming field is defined on the same grid as height_above_plane
        x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

        # Flatten the grid for interpolation
        points = np.c_[x.ravel(), y.ravel(), z.ravel()]

        # Extract the vector field
        if vector_field.shape == (1,self.xRes,self.yRes,self.zRes):
            n = 1
        elif vector_field.shape == (2,self.xRes,self.yRes,self.zRes):
            n = 2
        elif vector_field.shape == (3,self.xRes,self.yRes,self.zRes):
            n = 3
        else:
            raise Exception("You have entered an invalid field to the plot_vector_field function.")

        U = vector_field[0]
        U_values = U.ravel()
        U_verts = sp.interpolate.griddata(points, U_values, verts, method='nearest')

        if n > 1:
            V = vector_field[1]
            V_values = V.ravel()
            V_verts = sp.interpolate.griddata(points, V_values, verts, method='nearest')
        else:
            V_verts = np.zeros(U_verts.shape)
        
        if n > 2:
            W = vector_field[2]
            W_values = W.ravel()
            W_verts = sp.interpolate.griddata(points, W_values, verts, method='nearest')
        else:
            W_verts = np.zeros(U_verts.shape)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U_verts ** 2 + V_verts ** 2 + W_verts ** 2))
        U_verts = U_verts / max_vector
        V_verts = V_verts / max_vector
        W_verts = W_verts / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx)
        vy_scale = kwargs.get('vy_scale', spacing*self.dy)
        vz_scale = kwargs.get('vz_scale', spacing*self.dz)

        # Scaling
        U_verts = vx_scale*U_verts
        V_verts = vy_scale*V_verts
        W_verts = vz_scale*W_verts

        x = self.xmin+verts[:, 0]*self.dx
        y = self.ymin+verts[:, 1]*self.dy
        z = self.zmin+verts[:, 2]*self.dz

        # Adjust positions based on the coordinate system
        x = self.xmin + x * self.dx
        y = self.ymin + y * self.dy
        z = self.zmin + z * self.dz

        # Add spacing
        x = x[::spacing]
        y = y[::spacing]
        z = z[::spacing]
        U_verts = U_verts[::spacing]
        V_verts = V_verts[::spacing]
        W_verts = W_verts[::spacing]

        ax.quiver(x, y, z, U_verts, V_verts, W_verts, color='blue')

        kwargs['ax'] = ax
        self.plot_tool_set_axis_properties(**kwargs)
        return fig, ax


import numpy as np
from comfit.core.base_system import BaseSystem
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint


class PhaseFieldCrystal(BaseSystem):

    def __init__(self, dimension, **kwargs):
        """
        Initializes a system to simulate a Phase Field Crystal. This class is the base of the other
        phase field crystal models implemented in comfit

        Args:
            dimension: The dimension of the system.
            kwargs :keyword arguments to set additional parameters. See https://comfitlib.com/ClassPhaseFieldCrystal/

        Returns:
            The system object representing the phase field crystal simulation.
        """
        
        # First initialize the BaseSystem class
        super().__init__(dimension, **kwargs)

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.dislocation_charges = np.array(
            [[np.round(np.dot(an, qn) / (2 * np.pi), decimals=8) for qn in self.q] for an in self.a])

        # Elastic constant
        if ((self.dim - 1) * self.el_lambda + 2 * self.el_mu + self.el_gamma)!=0:
            self.el_nu = self.el_lambda / ((self.dim - 1) * self.el_lambda + 2 * self.el_mu + self.el_gamma)
        else:
            self.el_nu = 0

        self.Phi = 2*sum(np.array(self.eta0)**2)

    def __str__(self):
        """
        Returns a string representation of the object.

        Args:
            None
        
        Returns: 
            A string representation of the object.
        """
        string = "-------------------------------\n"
        string += "Phase Field Crystal instance\n"
        string += "-------------------------------\n"
        string += "Type: " + self.type + "\n"
        if self.dim == 1:
            string += "(nx) = (" + str(self.nx) + ")\n"
        elif self.dim == 2:
            string += "(nx, ny) = (" + str(self.nx) + ", " + str(self.ny) + ")\n"
        elif self.dim == 3:
            string += "(nx, ny, nz) = (" + str(self.nx) + ", " + str(self.ny) + ", " + str(self.nz) + ")\n"
        string += "micro_resolution: " + str(self.micro_resolution) + "\n"
        string += "-------------------------------\n"
        string += "Parameters\n"
        string += "-------------------------------\n"
        string += "r: " + str(self.r) + "\n"
        string += "t: " + str(self.t) + "\n"
        string += "v: " + str(self.v) + "\n"
        string += "-------------------------------\n"
        string += "Proto amplitudes\n"
        string += "-------------------------------\n"
        string += "psi0: " + str(self.psi0) + "\n"
        if self.type == 'PhaseFieldCrystal1DPeriodic':
            string += "A: " + str(self.A) + "\n"
        elif self.type == 'PhaseFieldCrystal2DTriangular':
            string += "A: " + str(self.A) + "\n"
        elif self.type == 'PhaseFieldCrystal2DSquare':
            string += "A: " + str(self.A) + "\n"
            string += "B: " + str(self.B) + "\n"
        elif self.type == 'PhaseFieldCrystal3DBodyCenteredCubic':
            string += "A: " + str(self.A) + "\n"
        elif self.type == 'PhaseFieldCrystal3DFaceCenteredCubic':
            string += "A: " + str(self.A) + "\n"
            string += "B: " + str(self.B) + "\n"
        elif self.type == 'PhaseFieldCrystal3DSimpleCubic':
            string += "A: " + str(self.A) + "\n"
            string += "B: " + str(self.B) + "\n"
            string += "C: " + str(self.C) + "\n"
        string += "-------------------------------\n"

        return string
    
    #######################################################
    ############ CONFIGURATION FUNCTIONS #################
    #######################################################

    def conf_PFC_from_amplitudes(self, eta=None):
        """Configures the PFC from the amplitudes.

        Args:
            eta: The amplitudes to configure the PFC from.

        Returns:
            Configures self.psi and self.psi_f.
        """

        self.psi = self.psi0

        if eta is None:
            eta = self.eta0

        for n in range(self.number_of_reciprocal_lattice_modes):

            if self.dim == 1:
                self.psi += 2 * eta[n] * np.exp(1j * self.q[n][0] * self.x)

            elif self.dim == 2:
                self.psi += 2 * eta[n] * np.exp(1j * (self.q[n][0] * self.x + self.q[n][1] * self.y))

            elif self.dim == 3:
                self.psi += 2 * eta[n] * np.exp(1j * (self.q[n][0] * self.x + self.q[n][1] * self.y + self.q[n][2] * self.z))

        self.psi = np.real(self.psi)
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_advect_PFC(self,u):
        """Advects the PFC according to the displacement field u.

        Args:
            u: The displacement field to advect the PFC with.

        Returns:
            None, but updates the PFC.
        """

        self.psi = np.real(self.calc_advect_field(self.psi, u, self.psi_f))
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_apply_strain(self, strain):
        """Applies a strain to the PFC.

        Args:
            strain: The strain to apply to the PFC.
        
        Returns:
            None, but updates the PFC.
        """

        if self.dim == 1:
            self.k[0] = self.k[0]/(1+strain)
            self.dif[0] = self.dif[0]/(1+strain)
            self.x = self.x*(1+strain)
        else:
            raise ImplementationError("Applied strain is not implemented for dimensions other than 1D.")

    def conf_create_polycrystal(self, type, **kwargs):
        """Creates a polycrystal.

        Args:
            type: The type of polycrystal to create.
            kwargs: Additional arguments for the polycrystal creation. Including:
                relaxation_time: The relaxation time to use for the polycrystal creation.
        
        Returns:
            None, but updates the PFC.
        """
        # First type of polycrystal, see documentation.
        if type == 'circular':
            # This creates a standard orientation of the crystal
            self.conf_PFC_from_amplitudes()

            rotation = kwargs.get('rotation', np.pi/6)
            position = kwargs.get('position', self.rmid)
            radius = kwargs.get('radius', self.Lmin/4)

            # Create the rotated field
            psi_rotated = self.calc_PFC_from_amplitudes(rotation=[0,0,rotation])

            # Set the rotated field in the inclusion region
            region  = self.calc_region_disk(position, radius)
            self.psi[region] = psi_rotated[region]
            self.psi_f = sp.fft.fftn(self.psi)

            # Smooth the interface
            relaxation_time = kwargs.get('relaxation_time', 10)
            self.evolve_PFC(round(relaxation_time/self.dt))

        elif type == 'four_grain':
            if self.dim == 1:
                raise Exception("Polycrystal type four_grain is not valid for 1 dimension.") 
                
            self.psi = self.calc_PFC_from_amplitudes(self.eta0)
            
            l1  = self.y>1/6*self.ymax+(4/6*self.ymax)/(1/1*self.xmax)*self.x
            l2  = self.y>1/2*self.ymax+(1/2*self.ymax)/(2/3*self.xmax)*self.x
            l3  = self.y>1/2*self.ymax-(1/2*self.ymax)/(2/3*self.xmax)*self.x
            l4  = self.y>5/6*self.ymax-(4/6*self.ymax)/(1/1*self.xmax)*self.x
            l5  = self.x>self.xmax/4
            l6  = self.y>5/4*self.ymax-(3/4*self.ymax)/(1/1*self.xmax)*self.x
            l7  = self.x>self.xmax/2
            l8  = self.y<1/2*self.ymax+(1/2*self.ymax)/(2/3*self.xmax)*self.x
            l9  = self.x>3/4*self.xmax
            l10 = self.y<-1/4*self.ymax+(3/4*self.ymax)/(1/1*self.xmax)*self.x

            zdir = 1 if self.dim == 2 else np.ones((1,1,self.zRes))

            pfcRotated = self.calc_PFC_from_amplitudes(self.eta0, rotation=[0,0,-22.5/180*np.pi])
            region = np.bool_((l4*~(l7)*~(l8) + ~(l1)*~(l3)*~(l7))*zdir)
            self.psi[region] = pfcRotated[region]

            pfcRotated = self.calc_PFC_from_amplitudes(self.eta0, rotation=[0,0,22.5/180*np.pi])
            region = np.bool_((l1*l6*l7 + l7*l10*~(l4))*zdir)
            self.psi[region] = pfcRotated[region]

            pfcRotated = self.calc_PFC_from_amplitudes(self.eta0, rotation=[0,0,45/180*np.pi])
            region = np.bool_((l1*~(l4)*~(l5) + ~(l1)*l4*l9)*zdir)
            self.psi[region] = pfcRotated[region]

            self.psi_f = sp.fft.fftn(self.psi)

            relaxation_time = kwargs.get('relaxation_time', 10)
            
            self.evolve_PFC(round(relaxation_time/self.dt))


    #######################################################
    ############### EVOLUTION FUNCTIONS ###################
    #######################################################

    def evolve_PFC(self, number_of_steps, method='ETD2RK'):
        """Evolves the PFC according to classical PFC dynamics.

        Args:
            number_of_steps: The number of steps to evolve the PFC.
            method: The method to use for the evolution (default: ETD2RK).
        
        Returns:
            Updates self.psi and self.psi_f
        """

        omega_f = self.calc_omega_f()

        integrating_factors_f, solver = self.calc_integrating_factors_f_and_solver(omega_f, method)

        for n in tqdm(range(number_of_steps), desc='Evolving the PFC'):
            self.psi, self.psi_f = solver(integrating_factors_f,
                                          self.calc_nonlinear_evolution_function_f,
                                          self.psi, self.psi_f)

            # These steps seem to be necessary for numerical stability (Vidar 18.12.23)
            self.psi = np.real(self.psi)
            self.psi_f = sp.fft.fftn(self.psi)

    def evolve_PFC_mechanical_equilibrium(self, time, Delta_t = 10, method='ETD2RK'):
        """Evolves the PFC in mechanical equilibrium. 

        Args:
            number_of_steps: The number of steps to evolve the PFC
            Delta_t: The time step for the mechanical equilibrium evolution
            method: The method to use for the evolution

        Returns:
            Updates self.psi and self.psi_f
        """
        
        number_of_steps = round(time/self.dt)
        number_of_steps_per_iteration = round(Delta_t/self.dt)
        number_of_iterations = round(number_of_steps/number_of_steps_per_iteration)

        for n in tqdm(range(number_of_iterations), desc='Evolving the PFC in mechanical equilibrium'):
            self.conf_advect_PFC(self.calc_displacement_field_to_equilibrium())
            self.evolve_PFC(number_of_steps_per_iteration, method)
            

    def evolve_PFC_hydrodynamic(self, number_of_steps, 
                                method = 'ETD2RK',
                                gamma_S = 2**-4,
                                rho0 = 2**-4):
        """Evolves the PFC according to hydrodynamic PFC dynamics.

        This requires introducing a velocity field. 
        If psi does not contain this field, it is added to the components psi[1],psi[2],psi[3].
        
        Args:
            number_of_steps: The number of steps to evolve the PFC.
            method: The method to use for the evolution (default: ETD2RK).
            gamma_S: The surface tension coefficient.
            rho0: The mass density.

        Returns:
            Updates self.psi and self.psi_f
        """
        
        if hasattr(self,'velocity_field'):
            pass
        else:
            self.velocity_field = True
            self.psi = np.array([self.psi]+[np.zeros_like(self.psi)]*self.dim)
            self.psi_f = np.array([self.psi_f]+[np.zeros_like(self.psi_f)]*self.dim)
            # print("psi shape", self.psi.shape)

            if not hasattr(self,'external_force_density_f'):
                self.external_force_density_f = np.zeros([self.dim] + self.dims, dtype=complex)

        self.gamma_S = gamma_S
        self.rho0 = rho0

        omega_f = self.calc_omega_hydrodynamic_f()
        integrating_factors_f, solver = self.calc_integrating_factors_f_and_solver(omega_f, method)

        for n in tqdm(range(number_of_steps), desc='Evolving the hydrodynamic PFC'):
            self.psi, self.psi_f = solver(integrating_factors_f,
                                          self.calc_nonlinear_hydrodynamic_evolution_function_f,
                                          self.psi, self.psi_f)
            
            self.psi = np.real(self.psi)
            self.psi_f = sp.fft.fftn(self.psi, axes = (range ( - self.dim , 0) ) )

    ############################################################
    ################# CALCULATION FUNCTIONS ####################
    ############################################################
    def calc_omega_hydrodynamic_f(self):
        """Calculates the hydrodynamic evolution function omega_f.

        Args:
            None

        Returns:
            The hydrodynamic evolution function omega_f.
        """
        k2 = self.calc_k2()
        return np.array([self.calc_omega_f()]+[-self.gamma_S/self.rho0*k2]*self.dim)


    def calc_PFC_free_energy_density_and_chemical_potential(self,field=None,field_f=None):
        """Calculates the free energy density and chemical potential of the PFC.

        Args:
            field: The field to calculate the free energy density and chemical potential of.
            field_f: The Fourier transform of the field.

        Returns:
            The free energy density and chemical potential of the PFC.
        """

        if field is None:
            field = self.psi
            field_f = self.psi_f

        psi_f = field_f
        
        # print("field shape",field.shape)
        # print("field_f shape",field_f.shape)

        psi = field 
        psi2 = field**2 
        psi3 = psi2*psi
        psi4 = psi2**2

        free_energy_density = 1/2*self.calc_Lpsi(psi_f)**2 \
            + 1/2*self.r*psi2 + 1/3*self.t*psi3 + 1/4*self.v*psi4
        
        # print("Lpsi shape",self.calc_Lpsi(psi_f).shape)

        chemical_potential = self.calc_L2psi(psi_f)*psi \
            + self.r*psi + self.t*psi2 + self.v*psi3
        
        # print("L2psi shape",self.calc_L2psi(psi_f).shape)
        
        # print("Free energy density shape", free_energy_density.shape)
        # print("Chem pot shape", chemical_potential.shape)
        return free_energy_density, chemical_potential

    def calc_nonlinear_hydrodynamic_evolution_function_f(self, field, t):
        """Calculates the hydrodynamic evolution function of the PFC.

        Args:
            field: The field to calculate the evolution function of.
            t: The time.

        Returns:
            The nonlinear evolution function for the hydrodynamic PFC.
        """

        field_f = sp.fft.fftn(field, axes =( range ( - self . dim , 0) ))

        k2 = self.calc_k2()

        N0_f = -k2*sp.fft.fftn(self.t * field[0] ** 2 + self.v * field[0] ** 3) \
            - sp.fft.fftn(sum([field[i+1]*sp.fft.ifftn(self.dif[i]*field_f[0]) for i in range(self.dim)]))
        
        force_density_f = self.calc_stress_divergence_f(field_f[0])

        return np.array([N0_f] + [1/self.rho0*(force_density_f[i]+self.external_force_density_f[i]) for i in range(self.dim)])

    def calc_nonlinear_evolution_function_f(self, psi, t):
        return -self.calc_k2()*sp.fft.fftn(self.t * psi ** 2 + self.v * psi ** 3)

    def calc_displacement_field_to_equilibrium(self):
        """Calculates the displacement field needed to put the PFC in mechanical equilibrium.

        Args:
            None
        
        Returns:
            The displacement field u
        """

        # Calculate the stress divergence
        g_f = self.calc_stress_divergence_f()

        # Calculate the displacement field
        u_f = np.zeros([self.dim] + self.dims, dtype = complex)

        k2 = self.calc_k2()
        k = np.sqrt(k2)

        kappa = np.zeros([self.dim] + self.dims)
        for l in range(self.dim):
            kappa[l] = self.k[l]/k

        second_term_factor = (self.el_mu + self.el_lambda)/(1+sum([(self.el_mu + self.el_lambda)/(self.el_mu + self.el_gamma*kappa[l]**2)*kappa[l]**2 for l in range(self.dim)]))

        # print(second_term_factor)

        for i in range(self.dim):
            for j in range(self.dim):
                #Calculate the Green's function
                Gij_f = int(i==j)/(self.el_mu + self.el_gamma*kappa[i]**2) \
                - kappa[i]*kappa[j]/((self.el_mu + self.el_gamma*kappa[i]**2)*(self.el_mu + self.el_gamma*kappa[j]**2))*second_term_factor

                u_f[i] += 1/k2*Gij_f*g_f[j]
            
            # Set the zero mode to zero
            u_f[i][self.zero_index] = 0

            # print(u_f[i])

        return np.real(sp.fft.ifftn(u_f, axes = (range ( - self . dim , 0) )))

        
    # Initial configuration methods
    def calc_amplitudes_with_dislocation(self, eta=None, x=None, y=None, dislocation_type=1):
        """ Calculates the amplitudes with a single point dislocation inserted.

        Args:
            eta: The amplitudes to insert the dislocation in.
            x: The x-coordinate of the dislocation.
            y: The y-coordinate of the dislocation.
            dislocation_type: The dislocation type to insert.
        
        Returns:
            The amplitudes containing the dislocation
        """

        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 for a single point dislocation.")

        if x == None:
            x = self.xmid
        if y == None:
            y = self.ymid
        if eta == None:
            eta = self.eta0

        sn = self.dislocation_charges[dislocation_type - 1]
        for n in range(self.number_of_reciprocal_lattice_modes):
            if sn[n] != 0:
                eta[n] *= np.exp(1j * self.calc_angle_field_single_vortex([x, y], charge=sn[n]))

        return eta

    def calc_amplitudes_with_dislocation_dipole(self, eta=None,
                                                x1=None, y1=None,
                                                x2=None, y2=None,
                                                dislocation_type=1):
        """
        Inserts a  dislocation dipole in the system corresponding to dislocation type and its negative

        Args:
            eta: The amplitudes to insert the dislocation dipole in
            x1: The x-coordinate of the first dislocation
            y1: The y-coordinate of the first dislocation
            x2: The x-coordinate of the second dislocation
            y2: The y-coordinate of the second dislocation
            dislocation_type: The dislocation type to insert
        
        Returns:
            The amplitudes with the dislocation dipole inserted
        """

        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 to insert a dislocation dipole.")

        if x1 == None:
            x1 = self.xmax / 3
        if y1 == None:
            y1 = self.ymax / 2
        if x2 == None:
            x2 = 2 * self.xmax / 3
        if y2 == None:
            y2 = self.ymax / 2

        if eta == None:
            eta = self.eta0.reshape(self.number_of_reciprocal_lattice_modes,1,1)\
                *np.ones([self.number_of_reciprocal_lattice_modes] + self.dims, dtype=complex)

        sn = self.dislocation_charges[dislocation_type - 1]

        theta = self.calc_angle_field_vortex_dipole(
            dipole_vector=[x2 - x1, y2 - y1],
            dipole_position=[(x1 + x2) / 2, (y1 + y2) / 2])

        for n in range(self.number_of_reciprocal_lattice_modes):
            if sn[n] != 0:
                eta[n] *= np.exp(1j * sn[n] * theta)

        return eta

    def calc_amplitudes_with_dislocation_ring(self, eta=None,
                                                position=None,
                                                radius=None,
                                                normal_vector=[0, 0, 1],
                                                dislocation_type=1):
        """
        Inserts a  dislocation ring in the system corresponding to dislocation type.

        Args:
            eta: The amplitudes to insert the dislocation ring in
            position: The position of the dislocation ring
            radius: The radius of the dislocation ring
            normal_vector: The normal vector of the dislocation ring
            dislocation_type: The dislocation type to insert

        Returns:
            The amplitudes with the dislocation ring inserted

        """

        if not (self.dim == 3):
            raise Exception("The dimension of the system must be 3 to insert a dislocation dipole.")

        if position == None:
            position = self.rmid

        if radius == None:
            radius = min(self.rmax)/3

        if eta == None:
            eta = self.eta0.reshape(self.number_of_reciprocal_lattice_modes,1,1,1)\
                *np.ones([self.number_of_reciprocal_lattice_modes] + self.dims, dtype=complex)

        sn = self.dislocation_charges[dislocation_type - 1]

        theta = self.calc_angle_field_vortex_ring(
            radius=radius,
            position=position,
            normal_vector=normal_vector)

        for n in range(self.number_of_reciprocal_lattice_modes):
            if sn[n] != 0:
                eta[n] *= np.exp(1j * sn[n] * theta)

        return eta
    
    def calc_PFC_from_amplitudes(self, eta=None, rotation=None):
        """Calculates the PFC from the amplitudes.

        Args:
            eta: The amplitudes to calculate the PFC from.
            rotation : The rotation of the PFC

        Returns:
            The PFC.
        """ 

        psi = self.psi0

        if rotation is None:
            rotation = [0,0,0]
        else:
            if isinstance(rotation, float):
                rotation = [0,0,rotation]

        rotation = sp.spatial.transform.Rotation.from_rotvec(rotation)
        rotation_matrix = rotation.as_matrix()

        # Rotate q-vectors to the new orientation
        if self.dim == 2:
            q = (rotation_matrix[0:2,0:2]@self.q.transpose()).transpose()
        elif self.dim == 3:
            q = (rotation_matrix[0:3,0:3]@self.q.transpose()).transpose()

        if eta is None:
            eta = self.eta0

        for n in range(self.number_of_reciprocal_lattice_modes):

            if self.dim == 1:
                psi += 2 * eta[n] * np.exp(1j * q[n][0] * self.x)

            elif self.dim == 2:
                psi += 2 * eta[n] * np.exp(1j * (q[n][0] * self.x + q[n][1] * self.y))

            elif self.dim == 3:
                psi += 2 * eta[n] * np.exp(1j * (q[n][0] * self.x + q[n][1] * self.y + q[n][2] * self.z))

        return np.real(psi)

    def calc_demodulate_PFC(self):
        """Demodulates the PFC.

        Args:
            None

        Returns:
            The amplitudes corresponding to the demodulated PFC.
        """

        eta = np.zeros([self.number_of_primary_reciprocal_lattice_modes] + self.dims, 
                       dtype=complex)

        Gaussian_filter_f = self.calc_Gaussian_filter_f()

        if hasattr(self,'velocity_field'):
            order_parameter = self.psi[0]
        else:
            order_parameter = self.psi

        if self.dim == 2:
                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    eta[n] = sp.fft.ifftn(Gaussian_filter_f*sp.fft.fftn(order_parameter*np.exp(
                        -1j*self.q[n][0]*self.x - 1j*self.q[n][1]*self.y)))

        elif self.dim == 3:
            for n in range(self.number_of_primary_reciprocal_lattice_modes):
                eta[n] = sp.fft.ifftn(Gaussian_filter_f*sp.fft.fftn(order_parameter*np.exp(
                    -1j*self.q[n][0]*self.x - 1j*self.q[n][1]*self.y - 1j*self.q[n][2]*self.z  
                    )))
                
        return eta

    def calc_stress_tensor(self):
        """Calculates the stress of the phase-field crystal.

        Args:
            None

        Returns:
            The stress tensor
        """
        
        stress = self.calc_stress_tensor_microscopic()
        Gaussian_filter_f = self.calc_Gaussian_filter_f()

        # Number of independent stress components
        if self.dim == 1:
            number_of_stress_components = 1
        elif self.dim == 2:
            number_of_stress_components = 3
        elif self.dim == 3:
            number_of_stress_components = 6

        # Coarse-grain the microscopic stress
        for n in range(number_of_stress_components):
            stress[n] = Gaussian_filter_f*sp.fft.fftn(stress[n])

        return np.real(sp.fft.ifftn(stress, axes = (range( - self.dim , 0) )))

    def calc_structure_tensor(self):
        """Calculates the structure tensor of the phase-field crystal.

        Args:
            None

        Returns:
            The structure tensor
        """
        # Calculate the gradient
        diPsi = np.zeros([self.dim] + self.dims, dtype=complex)
        for i in range(self.dim):
            diPsi[i] = self.dif[i]*self.psi_f
        diPsi = np.real(sp.fft.ifftn(diPsi, axes = (range( - self.dim , 0) )))

        if self.dim == 1:
            number_of_independent_strain_components = 1
        elif self.dim == 2:
            number_of_independent_strain_components = 3
        elif self.dim == 3:
            number_of_independent_strain_components = 6

        structure_tensor = np.zeros([number_of_independent_strain_components] + self.dims, dtype=complex)

        Gaussian_filter_f = self.calc_Gaussian_filter_f()
        if self.dim == 1:
            structure_tensor[0] = Gaussian_filter_f*sp.fft.fftn(diPsi[0]*diPsi[0])
        elif self.dim == 2:
            structure_tensor[0] = Gaussian_filter_f*sp.fft.fftn(diPsi[0]*diPsi[0])
            structure_tensor[1] = Gaussian_filter_f*sp.fft.fftn(diPsi[0]*diPsi[1])
            structure_tensor[2] = Gaussian_filter_f*sp.fft.fftn(diPsi[1]*diPsi[1])
        elif self.dim == 3:
            structure_tensor[0] = Gaussian_filter_f*sp.fft.fftn(diPsi[0]*diPsi[0])
            structure_tensor[1] = Gaussian_filter_f*sp.fft.fftn(diPsi[0]*diPsi[1])
            structure_tensor[2] = Gaussian_filter_f*sp.fft.fftn(diPsi[0]*diPsi[2])
            structure_tensor[3] = Gaussian_filter_f*sp.fft.fftn(diPsi[1]*diPsi[1])
            structure_tensor[4] = Gaussian_filter_f*sp.fft.fftn(diPsi[1]*diPsi[2])
            structure_tensor[5] = Gaussian_filter_f*sp.fft.fftn(diPsi[2]*diPsi[2])
        
        structure_tensor = np.real(sp.fft.ifftn(structure_tensor, axes = (range( - self.dim , 0) )))

        return structure_tensor
            


    def calc_strain_tensor(self):
        """Calculates the strain of the phase-field crystal.

        Args:
            None

        Returns:
            The strain tensor
        """
        
        strain = -self.dim/(2*self.Phi)*self.calc_structure_tensor()

        if self.dim == 1:
            strain = 1/2 + strain
        elif self.dim == 2:
            strain[0] = 1/2 + strain[0]
            strain[2] = 1/2 + strain[2]
        elif self.dim == 3:
            strain[0] = 1/2 + strain[0]
            strain[3] = 1/2 + strain[3]
            strain[5] = 1/2 + strain[5]
        
        return strain
        



    def calc_dislocation_density(self, eta = None):
        """Calculates the dislocation density

        Args:
            eta: The amplitudes to calculate the dislocation density from.

        Returns:
            The dislocation density
        """

        if eta is None:
            eta = self.calc_demodulate_PFC()

        if self.dim == 2:
            alpha = np.zeros([2] + self.dims)

            for n in range(self.number_of_primary_reciprocal_lattice_modes):
                D = self.calc_determinant_field([np.real(eta[n]), np.imag(eta[n])])
                alpha[0] += D*self.q[n,0]
                alpha[1] += D*self.q[n,1]

        elif self.dim == 3:
            alpha = np.zeros([3,3] + self.dims)

            for n in range(self.number_of_primary_reciprocal_lattice_modes):
                D = self.calc_determinant_field([np.real(eta[n]), np.imag(eta[n])])
                for i in range(3):
                    alpha[i][0] += D[i]*self.q[n,0]
                    alpha[i][1] += D[i]*self.q[n,1]
                    alpha[i][2] += D[i]*self.q[n,2]

        alpha = 2*self.dim/(self.number_of_primary_reciprocal_lattice_modes*self.A**2)*alpha

        return alpha
        

    def calc_dislocation_nodes(self):
        """ Calculates the dislocation nodes

        Args:
            None
        
        Returns:
            The dislocation nodes
        """

        alpha = self.calc_dislocation_density()

        if self.dim == 2:
            # self.plot_field(np.sqrt(alpha[0]**2 + alpha[1]**2))
            # plt.show()
            dislocation_nodes = self.calc_defect_nodes(np.sqrt(alpha[0]**2 + alpha[1]**2),
                                                       charge_tolerance=0.2*self.a0)
            
            for dislocation_node in dislocation_nodes:
                Burgers_vector = np.array([
                    alpha[0][dislocation_node['position_index']], 
                    alpha[1][dislocation_node['position_index']]
                    ])
                
                # print("Burgers vector:", Burgers_vector)
                
                # Find the Burgers vector 
                biggest_overlap = 0
                for a in self.a:
                    for sign in [-1, 1]:
                        overlap = np.dot(Burgers_vector, sign*a)
                        # print("Burgers vector",Burgers_vector)
                        # print("a-vector",a*sign)
                        # print("overlap",overlap)
                        # print("biggest overlap", biggest_overlap)
                        
                        if overlap > biggest_overlap:
                            # print("tjobing")
                            biggest_overlap = overlap
                            dislocation_node['Burgers_vector'] = sign*a
                
        elif self.dim == 3:
            dislocation_nodes = self.calc_defect_nodes(
                np.sqrt(
                      alpha[0][0]**2 + alpha[0][1]**2 + alpha[0][2]**2 \
                    + alpha[1][0]**2 + alpha[1][1]**2 + alpha[1][2]**2 \
                    + alpha[2][0]**2 + alpha[2][1]**2 + alpha[2][2]**2
                )
            )

            for dislocation_node in dislocation_nodes:
                alpha_tensor = np.array([
                    [alpha[0][0][dislocation_node['position_index']], alpha[0][1][dislocation_node['position_index']], alpha[0][2][dislocation_node['position_index']]],
                    [alpha[1][0][dislocation_node['position_index']], alpha[1][1][dislocation_node['position_index']], alpha[1][2][dislocation_node['position_index']]],
                    [alpha[2][0][dislocation_node['position_index']], alpha[2][1][dislocation_node['position_index']], alpha[2][2][dislocation_node['position_index']]]
                ])
                # print("alpha-tensor:",alpha_tensor)

                U, S, V = np.linalg.svd(alpha_tensor)
                tangent_vector = U[:,0]
                Burgers_vector = V[0,:]

                # print("U:",U)
                # print("S:",S)   
                # print("V:",V)

                # print("tangent vector:", tangent_vector)
                # print("Burgers vector:", Burgers_vector)

                # Find the Burgers vector 
                biggest_overlap = 0
                tangent_vector_sign = np.nan
                # print("Finding new Burgers vector...")
                for a in self.a:
                    for sign in [-1, 1]:
                        # print("Burgers vector",Burgers_vector)
                        # print("a-vector",a*sign)
                        # print("overlap",np.dot(Burgers_vector,sign*a))
                        overlap = np.dot(Burgers_vector, sign*a)
                        if overlap > biggest_overlap:
                            biggest_overlap = overlap
                            # print("New biggest overlap", overlap)
                            # print("Setting new BUrgers vector",a)
                            dislocation_node['Burgers_vector'] = a
                            tangent_vector_sign = sign

                dislocation_node['tangent_vector'] = tangent_vector_sign*tangent_vector

        return dislocation_nodes

    def calc_orientation_field(self):
        """Calculates the orientation field of the phase-field crystal.

        Args:
            None
        
        Returns:
            An orientation field, which is a vector field specifying the orientation of the crystal.
        """

        if hasattr(self,'velocity_field'):
            order_parameter = self.psi[0]
        else:
            order_parameter = self.psi

        eta = np.zeros([self.number_of_primary_reciprocal_lattice_modes] + self.dims, 
                       dtype=complex)
        Gaussian_filter_f = self.calc_Gaussian_filter_f()

        if self.dim == 2:
            resolution = 16
            orientation_field = 0 + 0j
            for angle in np.linspace(0, np.pi/3, resolution):
                rotation = sp.spatial.transform.Rotation.from_rotvec([0,0,angle])
                rotation_matrix = rotation.as_matrix()
                q = (rotation_matrix[0:2,0:2]@self.q.transpose()).transpose()
                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    eta[n] = sp.fft.ifftn(Gaussian_filter_f*sp.fft.fftn(order_parameter*np.exp(
                            -1j*q[n][0]*self.x - 1j*q[n][1]*self.y)))
                
                Phi2 = np.zeros(self.dims)
                for n in range(self.number_of_primary_reciprocal_lattice_modes):
                    Phi2 += abs(eta[n])**2
                
                orientation_field += np.exp(3*1j*angle)*Phi2

            return orientation_field



        pass


    #######################################################
    ############### PLOTTING FUNCTIONS ####################
    #######################################################

    def plot_dislocation_nodes(self, dislocation_nodes, **kwargs):
        """
        Plots the dislocation nodes.

        Args:
            dislocation_nodes: The dislocation nodes to plot.
            **kwargs: Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ for a full list of keyword arguments.  
                
        Returns: 
            The axes containing the plot. (matplotlib.axes.Axes)
        """    

        # Check if an axis object is provided
        fig = kwargs.get('fig', plt.gcf())
        ax = kwargs.get('ax', None)

        if self.dim == 2:

            if ax == None:
                fig.clf()
                ax = fig.add_subplot(111)

            x_coords = []
            y_coords = []

            # vx_coords = []
            # vy_coords = []

            Bx_coords = []
            By_coords = []

            for dislocation in dislocation_nodes:

                x_coords.append(dislocation['position'][0])
                y_coords.append(dislocation['position'][1])
                # vx_coords.append(vortex['velocity'][0])
                # vy_coords.append(vortex['velocity'][1])
                Bx_coords.append(dislocation['Burgers_vector'][0])
                By_coords.append(dislocation['Burgers_vector'][1])

            x_coords = np.array(x_coords)
            y_coords = np.array(y_coords)

            # print(x_coords_pos,y_coords_pos)
            # print(x_coords_neg,y_coords_neg)
            ax.scatter(x_coords/self.a0, y_coords/self.a0, marker='o', color='black')

            # ax.quiver(x_coords, y_coords, vx_coords, vy_coords, color='black')
            ax.quiver(x_coords/self.a0, y_coords/self.a0, Bx_coords, By_coords, color='red')
            
            kwargs['ax'] = ax
            self.plot_tool_set_axis_properties(**kwargs)

        elif self.dim == 3:
            # Plotting options
            quiver_scale = 2 # The scale of the quiver arrows

            if ax == None:
                plt.clf()
                ax = plt.gcf().add_subplot(111, projection='3d')

            x_coords = []
            y_coords = []
            z_coords = []

            tx = []
            ty = []
            tz = []

            # vx = []
            # vy = []
            # vz = []

            Bx = []
            By = []
            Bz = []


            for dislocation in dislocation_nodes:
                x_coords.append(dislocation['position'][0])
                y_coords.append(dislocation['position'][1])
                z_coords.append(dislocation['position'][2])

                tx.append(dislocation['tangent_vector'][0])
                ty.append(dislocation['tangent_vector'][1])
                tz.append(dislocation['tangent_vector'][2])

                # vx.append(dislocation['velocity'][0])
                # vy.append(dislocation['velocity'][1])
                # vz.append(dislocation['velocity'][2])

                Bx.append(dislocation['Burgers_vector'][0])
                By.append(dislocation['Burgers_vector'][1])
                Bz.append(dislocation['Burgers_vector'][2])

            tx = np.array(tx)
            ty = np.array(ty)
            tz = np.array(tz)

            # vx = np.array(vx)
            # vy = np.array(vy)
            # vz = np.array(vz)

            Bx = np.array(Bx)
            By = np.array(By)
            Bz = np.array(Bz)

            # if not len(vx) == 0:
            #     v2 =vx**2 + vy**2 + vz**2
            #     v_norm = np.sqrt(max(v2))
            # else:
            #     v_norm = 1

            if not len(Bx) == 0:
                B2 =Bx**2 + By**2 + Bz**2
                B_norm = np.sqrt(max(B2))
            else:
                B_norm = 1

            #ax.scatter(x_coords, y_coords, z_coords, marker='o', color='black')
            ax.quiver(x_coords, y_coords, z_coords, quiver_scale*tx, quiver_scale*ty, quiver_scale*tz, color='blue')
            # ax.quiver(x_coords, y_coords, z_coords, quiver_scale*vx/v_norm, quiver_scale*vy/v_norm, quiver_scale*vz/v_norm, color='green')
            ax.quiver(x_coords, y_coords, z_coords, quiver_scale*Bx/B_norm, quiver_scale*By/B_norm, quiver_scale*Bz/B_norm, color='red')

            ax.set_xlabel('$x/a_0$')
            ax.set_ylabel('$y/a_0$')
            ax.set_zlabel('$z/a_0$')

            ax.set_xlim([0, self.xmax-self.dx])
            ax.set_ylim([0, self.ymax-self.dy])
            ax.set_zlim([0, self.zmax-self.dz])

            ax.set_aspect('equal')